"""Despite being public API, this is only meant for use in our certbot-nginx plugin, and isn't
intended for public use."""

from certbot._internal.plugins.nginx import configurator


ENTRYPOINT = configurator.NginxConfigurator



"""Decide which plugins to use for authentication & installation"""

import logging
from typing import cast
from typing import Iterable
from typing import Optional
from typing import TypeVar

from certbot import configuration
from certbot import errors
from certbot import interfaces
from certbot._internal.plugins import disco
from certbot.compat import os
from certbot.display import util as display_util

logger = logging.getLogger(__name__)


def pick_configurator(config: configuration.NamespaceConfig, default: Optional[str],
                      plugins: disco.PluginsRegistry,
                      question: str = "How would you like to authenticate and install "
                                      "certificates?") -> Optional[interfaces.Plugin]:
    """Pick configurator plugin."""
    return pick_plugin(
        config, default, plugins, question,
        (interfaces.Authenticator, interfaces.Installer))


def pick_installer(config: configuration.NamespaceConfig, default: Optional[str],
                   plugins: disco.PluginsRegistry,
                   question: str = "How would you like to install certificates?"
                   ) -> Optional[interfaces.Installer]:
    """Pick installer plugin."""
    return pick_plugin(config, default, plugins, question, (interfaces.Installer,))


def pick_authenticator(config: configuration.NamespaceConfig, default: Optional[str],
                       plugins: disco.PluginsRegistry,
                       question: str = "How would you "
                                       "like to authenticate with the ACME CA?"
                       ) -> Optional[interfaces.Authenticator]:
    """Pick authentication plugin."""
    return pick_plugin(
        config, default, plugins, question, (interfaces.Authenticator,))


def get_unprepared_installer(config: configuration.NamespaceConfig,
                             plugins: disco.PluginsRegistry) -> Optional[interfaces.Installer]:
    """
    Get an unprepared interfaces.Installer object.

    :param certbot.configuration.NamespaceConfig config: Configuration
    :param certbot._internal.plugins.disco.PluginsRegistry plugins:
        All plugins registered as entry points.

    :returns: Unprepared installer plugin or None
    :rtype: Plugin or None
    """

    _, req_inst = cli_plugin_requests(config)
    if not req_inst:
        return None
    installers = plugins.filter(lambda p_ep: p_ep.check_name(req_inst))
    installers.init(config)
    if len(installers) > 1:
        raise errors.PluginSelectionError(
            "Found multiple installers with the name %s, Certbot is unable to "
            "determine which one to use. Skipping." % req_inst)
    if installers:
        inst = list(installers.values())[0]
        logger.debug("Selecting plugin: %s", inst)
        return inst.init(config)
    raise errors.PluginSelectionError(
        "Could not select or initialize the requested installer %s." % req_inst)


P = TypeVar('P', bound=interfaces.Plugin)


def pick_plugin(config: configuration.NamespaceConfig, default: Optional[str],
                plugins: disco.PluginsRegistry, question: str,
                ifaces: Iterable[type]) -> Optional[P]:
    """Pick plugin.

    :param certbot.configuration.NamespaceConfig config: Configuration
    :param str default: Plugin name supplied by user or ``None``.
    :param certbot._internal.plugins.disco.PluginsRegistry plugins:
        All plugins registered as entry points.
    :param str question: Question to be presented to the user in case
        multiple candidates are found.
    :param list ifaces: Interfaces that plugins must provide.

    :returns: Initialized plugin.
    :rtype: Plugin

    """
    if default is not None:
        # throw more UX-friendly error if default not in plugins
        filtered = plugins.filter(lambda p_ep: p_ep.check_name(default))
    else:
        if config.noninteractive_mode:
            # it's really bad to auto-select the single available plugin in
            # non-interactive mode, because an update could later add a second
            # available plugin
            raise errors.MissingCommandlineFlag(
                "Missing command line flags. For non-interactive execution, "
                "you will need to specify a plugin on the command line.  Run "
                "with '--help plugins' to see a list of options, and see "
                "https://eff.org/letsencrypt-plugins for more detail on what "
                "the plugins do and how to use them.")

        filtered = plugins.visible()

    filtered = filtered.ifaces(ifaces)
    filtered.init(config)
    filtered.prepare()
    prepared = filtered.available()

    if len(prepared) > 1:
        logger.debug("Multiple candidate plugins: %s", prepared)
        plugin_ep1 = choose_plugin(list(prepared.values()), question)
        if plugin_ep1 is None:
            return None
        return cast(P, plugin_ep1.init())
    elif len(prepared) == 1:
        plugin_ep2 = list(prepared.values())[0]
        logger.debug("Single candidate plugin: %s", plugin_ep2)
        if plugin_ep2.misconfigured:
            return None
        return plugin_ep2.init()
    else:
        logger.debug("No candidate plugin")
        return None


def choose_plugin(prepared: list[disco.PluginEntryPoint],
                  question: str) -> Optional[disco.PluginEntryPoint]:
    """Allow the user to choose their plugin.

    :param list prepared: List of `~.PluginEntryPoint`.
    :param str question: Question to be presented to the user.

    :returns: Plugin entry point chosen by the user.
    :rtype: `~.PluginEntryPoint`

    """
    opts = [plugin_ep.description_with_name +
            (" [Misconfigured]" if plugin_ep.misconfigured else "")
            for plugin_ep in prepared]

    while True:
        code, index = display_util.menu(question, opts, force_interactive=True)

        if code == display_util.OK:
            plugin_ep = prepared[index]
            if plugin_ep.misconfigured:
                display_util.notification(
                    "The selected plugin encountered an error while parsing "
                    "your server configuration and cannot be used. The error "
                    "was:\n\n{0}".format(plugin_ep.prepare()), pause=False)
            else:
                return plugin_ep
        else:
            return None


noninstaller_plugins = ["webroot", "manual", "standalone", "dns-cloudflare",
                        "dns-digitalocean", "dns-dnsimple", "dns-dnsmadeeasy", "dns-gehirn",
                        "dns-google", "dns-linode", "dns-luadns", "dns-nsone", "dns-ovh",
                        "dns-rfc2136", "dns-route53", "dns-sakuracloud"]


def record_chosen_plugins(config: configuration.NamespaceConfig, plugins: disco.PluginsRegistry,
                          auth: Optional[interfaces.Authenticator],
                          inst: Optional[interfaces.Installer]) -> None:
    """Update the config entries to reflect the plugins we actually selected."""
    config.authenticator = None
    if auth:
        auth_ep = plugins.find_init(auth)
        if auth_ep:
            config.authenticator = auth_ep.name
    config.installer = None
    if inst:
        inst_ep = plugins.find_init(inst)
        if inst_ep:
            config.installer = inst_ep.name
    logger.info("Plugins selected: Authenticator %s, Installer %s",
                config.authenticator, config.installer)


def choose_configurator_plugins(config: configuration.NamespaceConfig,
                                plugins: disco.PluginsRegistry,
                                verb: str) -> tuple[Optional[interfaces.Installer],
                                                    Optional[interfaces.Authenticator]]:
    """
    Figure out which configurator we're going to use, modifies
    config.authenticator and config.installer strings to reflect that choice if
    necessary.

    :raises errors.PluginSelectionError if there was a problem

    :returns: tuple of (`Installer` or None, `Authenticator` or None)
    :rtype: tuple
    """

    req_auth, req_inst = cli_plugin_requests(config)
    installer_question = ""

    if verb == "enhance":
        installer_question = ("Which installer would you like to use to "
                              "configure the selected enhancements?")

    # Which plugins do we need?
    if verb == "run":
        need_inst = need_auth = True
        from certbot._internal.cli import cli_command
        if req_auth in noninstaller_plugins and not req_inst:
            msg = ('With the {0} plugin, you probably want to use the "certonly" command, eg:{1}'
                   '{1}    {2} certonly --{0}{1}{1}'
                   '(Alternatively, add a --installer flag. See https://eff.org/letsencrypt-plugins'
                   '{1} and "--help plugins" for more information.)'.format(
                       req_auth, os.linesep, cli_command))

            raise errors.MissingCommandlineFlag(msg)
    else:
        need_inst = need_auth = False
    if verb == "certonly":
        need_auth = True
    elif verb in ("install", "enhance"):
        need_inst = True
        if config.authenticator:
            logger.warning("Specifying an authenticator doesn't make sense when "
                           "running Certbot with verb \"%s\"", verb)
    # Try to meet the user's request and/or ask them to pick plugins
    authenticator: Optional[interfaces.Authenticator] = None
    installer: Optional[interfaces.Installer] = None
    if verb == "run" and req_auth == req_inst:
        # Unless the user has explicitly asked for different auth/install,
        # only consider offering a single choice
        configurator = pick_configurator(config, req_inst, plugins)
        authenticator = cast(Optional[interfaces.Authenticator], configurator)
        installer = cast(Optional[interfaces.Installer], configurator)
    else:
        if need_inst or req_inst:
            installer = pick_installer(config, req_inst, plugins, installer_question)
        if need_auth:
            authenticator = pick_authenticator(config, req_auth, plugins)

    # Report on any failures
    if need_inst and not installer:
        diagnose_configurator_problem("installer", req_inst, plugins)
    if need_auth and not authenticator:
        diagnose_configurator_problem("authenticator", req_auth, plugins)

    # As a special case for certonly, if a user selected apache or nginx, set
    # the relevant installer (unless the user specifically specified no
    # installer or only specified an authenticator on the command line)
    if verb == "certonly" and authenticator is not None:
        # user specified --nginx or --apache on CLI
        selected_configurator = config.nginx or config.apache
        # user didn't request an authenticator, and so interactively chose nginx
        # or apache
        interactively_selected = req_auth is None and authenticator.name in ("nginx", "apache")

        if selected_configurator or interactively_selected:
            installer = cast(Optional[interfaces.Installer], authenticator)
    logger.debug("Selected authenticator %s and installer %s", authenticator, installer)

    record_chosen_plugins(config, plugins, authenticator, installer)
    return installer, authenticator


def set_configurator(previously: Optional[str], now: Optional[str]) -> Optional[str]:
    """
    Setting configurators multiple ways is okay, as long as they all agree
    :param str previously: previously identified request for the installer/authenticator
    :param str now: the request currently being processed
    """
    if not now:
        # we're not actually setting anything
        return previously
    if previously:
        if previously != now:
            msg = "Too many flags setting configurators/installers/authenticators {0} -> {1}"
            raise errors.PluginSelectionError(msg.format(repr(previously), repr(now)))
    return now


def cli_plugin_requests(config: configuration.NamespaceConfig
                        ) -> tuple[Optional[str], Optional[str]]:
    """
    Figure out which plugins the user requested with CLI and config options

    :returns: (requested authenticator string or None, requested installer string or None)
    :rtype: tuple
    """
    req_inst = req_auth = config.configurator
    req_inst = set_configurator(req_inst, config.installer)
    req_auth = set_configurator(req_auth, config.authenticator)

    if config.nginx:
        req_inst = set_configurator(req_inst, "nginx")
        req_auth = set_configurator(req_auth, "nginx")
    if config.apache:
        req_inst = set_configurator(req_inst, "apache")
        req_auth = set_configurator(req_auth, "apache")
    if config.standalone:
        req_auth = set_configurator(req_auth, "standalone")
    if config.webroot:
        req_auth = set_configurator(req_auth, "webroot")
    if config.manual:
        req_auth = set_configurator(req_auth, "manual")
    if config.dns_cloudflare:
        req_auth = set_configurator(req_auth, "dns-cloudflare")
    if config.dns_digitalocean:
        req_auth = set_configurator(req_auth, "dns-digitalocean")
    if config.dns_dnsimple:
        req_auth = set_configurator(req_auth, "dns-dnsimple")
    if config.dns_dnsmadeeasy:
        req_auth = set_configurator(req_auth, "dns-dnsmadeeasy")
    if config.dns_gehirn:
        req_auth = set_configurator(req_auth, "dns-gehirn")
    if config.dns_google:
        req_auth = set_configurator(req_auth, "dns-google")
    if config.dns_linode:
        req_auth = set_configurator(req_auth, "dns-linode")
    if config.dns_luadns:
        req_auth = set_configurator(req_auth, "dns-luadns")
    if config.dns_nsone:
        req_auth = set_configurator(req_auth, "dns-nsone")
    if config.dns_ovh:
        req_auth = set_configurator(req_auth, "dns-ovh")
    if config.dns_rfc2136:
        req_auth = set_configurator(req_auth, "dns-rfc2136")
    if config.dns_route53:
        req_auth = set_configurator(req_auth, "dns-route53")
    if config.dns_sakuracloud:
        req_auth = set_configurator(req_auth, "dns-sakuracloud")
    logger.debug("Requested authenticator %s and installer %s", req_auth, req_inst)
    return req_auth, req_inst


def diagnose_configurator_problem(cfg_type: str, requested: Optional[str],
                                  plugins: disco.PluginsRegistry) -> None:
    """
    Raise the most helpful error message about a plugin being unavailable

    :param str cfg_type: either "installer" or "authenticator"
    :param str requested: the plugin that was requested
    :param .PluginsRegistry plugins: available plugins

    :raises error.PluginSelectionError: if there was a problem
    """

    if requested:
        if requested not in plugins:
            msg = "The requested {0} plugin does not appear to be installed".format(requested)
        else:
            msg = ("The {0} plugin is not working; there may be problems with "
                   "your existing configuration.\nThe error was: {1!r}"
                   .format(requested, plugins[requested].problem))
    elif cfg_type == "installer":
        from certbot._internal.cli import cli_command
        msg = ('Certbot doesn\'t know how to automatically configure the web '
          'server on this system. However, it can still get a certificate for '
          'you. Please run "{0} certonly" to do so. You\'ll need to '
          'manually configure your web server to use the resulting '
          'certificate.').format(cli_command)
    else:
        msg = "{0} could not be determined or is not installed".format(cfg_type)
    raise errors.PluginSelectionError(msg)



"""Tests for certbot._internal.plugins.apache.constants"""
import sys
import os.path

import pytest

from certbot._internal.plugins.apache import constants


def test_augeas_lens_dir_exists():
    assert os.path.exists(constants.AUGEAS_LENS_DIR)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Certbot command line constants"""
cli_command = "certbot"


# Argparse's help formatting has a lot of unhelpful peculiarities, so we want
# to replace as much of it as we can...

# This is the stub to include in help generated by argparse
SHORT_USAGE = """
  {0} [SUBCOMMAND] [options] [-d DOMAIN] [-d DOMAIN] ...

Certbot can obtain and install HTTPS/TLS/SSL certificates.  By default,
it will attempt to use a webserver both for obtaining and installing the
certificate. """.format(cli_command)

# This section is used for --help and --help all ; it needs information
# about installed plugins to be fully formatted
COMMAND_OVERVIEW = """The most common SUBCOMMANDS and flags are:

obtain, install, and renew certificates:
    (default) run   Obtain & install a certificate in your current webserver
    certonly        Obtain or renew a certificate, but do not install it
    renew           Renew all previously obtained certificates that are near expiry
    enhance         Add security enhancements to your existing configuration
   -d DOMAINS       Comma-separated list of domains to obtain a certificate for

  %s
  --standalone      Run a standalone webserver for authentication
  %s
  --webroot         Place files in a server's webroot folder for authentication
  --manual          Obtain certificates interactively, or using shell script hooks

   -n               Run non-interactively
  --test-cert       Obtain a test certificate from a staging server
  --dry-run         Test "renew" or "certonly" without saving any certificates to disk

manage certificates:
    certificates    Display information about certificates you have from Certbot
    revoke          Revoke a certificate (supply --cert-name or --cert-path)
    delete          Delete a certificate (supply --cert-name)
    reconfigure     Update a certificate's configuration (supply --cert-name)

manage your account:
    register        Create an ACME account
    unregister      Deactivate an ACME account
    update_account  Update an ACME account
    show_account    Display account details
  --agree-tos       Agree to the ACME server's Subscriber Agreement
   -m EMAIL         Email address for important account notifications
"""

# This is the short help for certbot --help, where we disable argparse
# altogether
HELP_AND_VERSION_USAGE = """
More detailed help:

  -h, --help [TOPIC]    print this message, or detailed help on a topic;
                        the available TOPICS are:

   all, automation, commands, paths, security, testing, or any of the
   subcommands or plugins (certonly, renew, install, register, nginx,
   apache, standalone, webroot, etc.)
  -h all                print a detailed help page including all topics
  --version             print the version number
"""

# These argparse parameters should be removed when detecting defaults.
ARGPARSE_PARAMS_TO_REMOVE = ("const", "nargs", "type",)


# These sets are used when to help detect options set by the user.
EXIT_ACTIONS = {"help", "version",}


ZERO_ARG_ACTIONS = {"store_const", "store_true",
                        "store_false", "append_const", "count",}


# Maps a config option to a set of config options that may have modified it.
# This dictionary is used recursively, so if A modifies B and B modifies C,
# it is determined that C was modified by the user if A was modified.
VAR_MODIFIERS = {"account": {"server",},
                 "server": {"dry_run", "staging",},
                 "webroot_map": {"webroot_path",}}

# This is a list of all CLI options that we have ever deprecated. It lets us
# opt out of the default detection, which can interact strangely with option
# deprecation. See https://github.com/certbot/certbot/issues/8540 for more info.
DEPRECATED_OPTIONS = {
    "manual_public_ip_logging_ok",
    "os_packages_only",
    "no_self_upgrade",
    "no_bootstrap",
    "no_permissions_check",
    "dns_route53_propagation_seconds",
    "certbot_route53:auth_propagation_seconds"
}



"""Common test utilities for the apache plugin."""
import shutil
from typing import Optional
import unittest
from unittest import mock

import augeas
import josepy as jose

from certbot import util
from certbot.compat import os
from certbot.plugins import common
from certbot.tests import util as test_util
from certbot._internal.plugins.apache import configurator
from certbot._internal.plugins.apache import entrypoint
from certbot._internal.plugins.apache import obj


class ApacheTest(unittest.TestCase):

    def setUp(self, test_dir: str = "debian_apache_2_4/multiple_vhosts",
              config_root: str = "debian_apache_2_4/multiple_vhosts/apache2",
              vhost_root: str = "debian_apache_2_4/multiple_vhosts/apache2/sites-available"
              ) -> None:
        # pylint: disable=arguments-differ
        self.temp_dir, self.config_dir, self.work_dir = common.dir_setup(
            test_dir=test_dir,
            pkg=__package__)

        self.config_path = os.path.join(self.temp_dir, config_root)
        self.vhost_path = os.path.join(self.temp_dir, vhost_root)

        self.rsa512jwk = jose.jwk.JWKRSA.load(test_util.load_vector(
            "rsa512_key.pem"))

        self.config = get_apache_configurator(self.config_path, vhost_root,
                                              self.config_dir, self.work_dir)

        # Make sure all vhosts in sites-enabled are symlinks (Python packaging
        # does not preserve symlinks)
        sites_enabled = os.path.join(self.config_path, "sites-enabled")
        if not os.path.exists(sites_enabled):
            return

        for vhost_basename in os.listdir(sites_enabled):
            # Keep the one non-symlink test vhost in place
            if vhost_basename == "non-symlink.conf":
                continue
            vhost = os.path.join(sites_enabled, vhost_basename)
            if not os.path.islink(vhost):  # pragma: no cover
                os.remove(vhost)
                target = os.path.join(
                    os.path.pardir, "sites-available", vhost_basename)
                os.symlink(target, vhost)

    def tearDown(self) -> None:
        util._release_locks()
        shutil.rmtree(self.temp_dir)
        shutil.rmtree(self.config_dir)
        shutil.rmtree(self.work_dir)


class ParserTest(ApacheTest):

    def setUp(self, test_dir: str = "debian_apache_2_4/multiple_vhosts",
              config_root: str = "debian_apache_2_4/multiple_vhosts/apache2",
              vhost_root: str = "debian_apache_2_4/multiple_vhosts/apache2/sites-available"
              ) -> None:
        super().setUp(test_dir, config_root, vhost_root)

        from certbot._internal.plugins.apache.parser import ApacheParser
        self.aug = augeas.Augeas(
            flags=augeas.Augeas.NONE | augeas.Augeas.NO_MODL_AUTOLOAD)
        with mock.patch("certbot._internal.plugins.apache.parser.ApacheParser."
                        "update_runtime_variables"):
            self.parser = ApacheParser(
                self.config_path, self.config, self.vhost_path)


def get_apache_configurator(
        config_path: str, vhost_path: str,
        config_dir: str, work_dir: str, version: tuple[int, int, int] = (2, 4, 7),
        os_info: str = "generic",
        conf_vhost_path: Optional[str] = None,
        use_parsernode: bool = False,
        openssl_version: str = "1.1.1a") -> configurator.ApacheConfigurator:
    """Create an Apache Configurator with the specified options.

    :param conf: Function that returns binary paths. self.conf in Configurator

    """
    backups = os.path.join(work_dir, "backups")
    mock_le_config = mock.MagicMock(
        apache_server_root=config_path,
        apache_vhost_root=None,
        apache_le_vhost_ext="-le-ssl.conf",
        apache_challenge_location=config_path,
        apache_enmod=None,
        backup_dir=backups,
        config_dir=config_dir,
        http01_port=80,
        temp_checkpoint_dir=os.path.join(work_dir, "temp_checkpoints"),
        in_progress_dir=os.path.join(backups, "IN_PROGRESS"),
        work_dir=work_dir)

    with mock.patch("certbot._internal.plugins.apache.configurator.util.run_script"):
        with mock.patch("certbot._internal.plugins.apache.configurator.util."
                        "exe_exists") as mock_exe_exists:
            mock_exe_exists.return_value = True
            with mock.patch("certbot._internal.plugins.apache.parser.ApacheParser."
                            "update_runtime_variables"):
                with mock.patch("certbot._internal.plugins.apache.apache_util.parse_from_subprocess") as mock_sp:
                    mock_sp.return_value = []
                    try:
                        config_class = entrypoint.OVERRIDE_CLASSES[os_info]
                    except KeyError:
                        config_class = configurator.ApacheConfigurator
                    config = config_class(config=mock_le_config, name="apache",
                                          version=version, use_parsernode=use_parsernode,
                                          openssl_version=openssl_version)
                    if not conf_vhost_path:
                        config_class.OS_DEFAULTS.vhost_root = vhost_path
                    else:
                        # Custom virtualhost path was requested
                        config.config.apache_vhost_root = conf_vhost_path
                    config.config.apache_ctl = config_class.OS_DEFAULTS.ctl
                    config.config.apache_bin = config_class.OS_DEFAULTS.bin
                    config.prepare()
    return config


def get_vh_truth(temp_dir: str, config_name: str) -> Optional[list[obj.VirtualHost]]:
    """Return the ground truth for the specified directory."""
    if config_name == "debian_apache_2_4/multiple_vhosts":
        prefix = os.path.join(
            temp_dir, config_name, "apache2/sites-enabled")

        aug_pre = "/files" + prefix
        vh_truth = [
            obj.VirtualHost(
                os.path.join(prefix, "encryption-example.conf"),
                os.path.join(aug_pre, "encryption-example.conf/Virtualhost"),
                {obj.Addr.fromstring("*:80")},
                False, True, "encryption-example.demo"),
            obj.VirtualHost(
                os.path.join(prefix, "default-ssl.conf"),
                os.path.join(aug_pre,
                             "default-ssl.conf/IfModule/VirtualHost"),
                {obj.Addr.fromstring("_default_:443")}, True, True),
            obj.VirtualHost(
                os.path.join(prefix, "000-default.conf"),
                os.path.join(aug_pre, "000-default.conf/VirtualHost"),
                {obj.Addr.fromstring("*:80"),
                    obj.Addr.fromstring("[::]:80")},
                False, True, "ip-172-30-0-17"),
            obj.VirtualHost(
                os.path.join(prefix, "certbot.conf"),
                os.path.join(aug_pre, "certbot.conf/VirtualHost"),
                {obj.Addr.fromstring("*:80")}, False, True,
                "certbot.demo", aliases={"www.certbot.demo"}),
            obj.VirtualHost(
                os.path.join(prefix, "mod_macro-example.conf"),
                os.path.join(aug_pre,
                             "mod_macro-example.conf/Macro/VirtualHost"),
                {obj.Addr.fromstring("*:80")}, False, True,
                modmacro=True),
            obj.VirtualHost(
                os.path.join(prefix, "default-ssl-port-only.conf"),
                os.path.join(aug_pre, ("default-ssl-port-only.conf/"
                                       "IfModule/VirtualHost")),
                {obj.Addr.fromstring("_default_:443")}, True, True),
            obj.VirtualHost(
                os.path.join(prefix, "wildcard.conf"),
                os.path.join(aug_pre, "wildcard.conf/VirtualHost"),
                {obj.Addr.fromstring("*:80")}, False, True,
                "ip-172-30-0-17", aliases={"*.blue.purple.com"}),
            obj.VirtualHost(
                os.path.join(prefix, "ocsp-ssl.conf"),
                os.path.join(aug_pre, "ocsp-ssl.conf/IfModule/VirtualHost"),
                {obj.Addr.fromstring("10.2.3.4:443")}, True, True,
                "ocspvhost.com"),
            obj.VirtualHost(
                os.path.join(prefix, "non-symlink.conf"),
                os.path.join(aug_pre, "non-symlink.conf/VirtualHost"),
                {obj.Addr.fromstring("*:80")}, False, True,
                "nonsym.link"),
            obj.VirtualHost(
                os.path.join(prefix, "default-ssl-port-only.conf"),
                os.path.join(aug_pre,
                             "default-ssl-port-only.conf/VirtualHost"),
                {obj.Addr.fromstring("*:80")}, True, True, ""),
            obj.VirtualHost(
                os.path.join(temp_dir, config_name,
                             "apache2/apache2.conf"),
                "/files" + os.path.join(temp_dir, config_name,
                                        "apache2/apache2.conf/VirtualHost"),
                {obj.Addr.fromstring("*:80")}, False, True,
                "vhost.in.rootconf"),
            obj.VirtualHost(
                os.path.join(prefix, "duplicatehttp.conf"),
                os.path.join(aug_pre, "duplicatehttp.conf/VirtualHost"),
                {obj.Addr.fromstring("10.2.3.4:80")}, False, True,
                "duplicate.example.com"),
            obj.VirtualHost(
                os.path.join(prefix, "duplicatehttps.conf"),
                os.path.join(aug_pre, "duplicatehttps.conf/IfModule/VirtualHost"),
                {obj.Addr.fromstring("10.2.3.4:443")}, True, True,
                "duplicate.example.com")]
        return vh_truth
    if config_name == "debian_apache_2_4/multi_vhosts":
        prefix = os.path.join(
            temp_dir, config_name, "apache2/sites-available")
        aug_pre = "/files" + prefix
        vh_truth = [
            obj.VirtualHost(
                os.path.join(prefix, "default.conf"),
                os.path.join(aug_pre, "default.conf/VirtualHost[1]"),
                {obj.Addr.fromstring("*:80")},
                False, True, "ip-172-30-0-17"),
            obj.VirtualHost(
                os.path.join(prefix, "default.conf"),
                os.path.join(aug_pre, "default.conf/VirtualHost[2]"),
                {obj.Addr.fromstring("*:80")},
                False, True, "banana.vomit.com"),
            obj.VirtualHost(
                os.path.join(prefix, "multi-vhost.conf"),
                os.path.join(aug_pre, "multi-vhost.conf/VirtualHost[1]"),
                {obj.Addr.fromstring("*:80")},
                False, True, "1.multi.vhost.tld"),
            obj.VirtualHost(
                os.path.join(prefix, "multi-vhost.conf"),
                os.path.join(aug_pre, "multi-vhost.conf/IfModule/VirtualHost"),
                {obj.Addr.fromstring("*:80")},
                False, True, "2.multi.vhost.tld"),
            obj.VirtualHost(
                os.path.join(prefix, "multi-vhost.conf"),
                os.path.join(aug_pre, "multi-vhost.conf/VirtualHost[2]"),
                {obj.Addr.fromstring("*:80")},
                False, True, "3.multi.vhost.tld")]
        return vh_truth
    return None  # pragma: no cover



""" Distribution specific override class for Debian family (Ubuntu/Debian) """
import logging

from certbot import errors
from certbot import util
from certbot.compat import filesystem
from certbot.compat import os
from certbot._internal.plugins.apache import apache_util
from certbot._internal.plugins.apache import configurator
from certbot._internal.plugins.apache.configurator import OsOptions
from certbot._internal.plugins.apache.obj import VirtualHost

logger = logging.getLogger(__name__)


class DebianConfigurator(configurator.ApacheConfigurator):
    """Debian specific ApacheConfigurator override class"""

    OS_DEFAULTS = OsOptions(
        enmod="a2enmod",
        dismod="a2dismod",
        handle_modules=True,
        handle_sites=True,
    )

    def enable_site(self, vhost: VirtualHost) -> None:
        """Enables an available site, Apache reload required.

        .. note:: Does not make sure that the site correctly works or that all
                  modules are enabled appropriately.

        :param vhost: vhost to enable
        :type vhost: :class:`~certbot._internal.plugins.apache.obj.VirtualHost`

        :raises .errors.NotSupportedError: If filesystem layout is not
            supported.

        """
        if vhost.enabled:
            return None

        enabled_path = ("%s/sites-enabled/%s" %
                        (self.parser.root,
                         os.path.basename(vhost.filep)))
        if not os.path.isdir(os.path.dirname(enabled_path)):
            # For some reason, sites-enabled / sites-available do not exist
            # Call the parent method
            return super().enable_site(vhost)
        self.reverter.register_file_creation(False, enabled_path)
        try:
            os.symlink(vhost.filep, enabled_path)
        except OSError as err:
            if os.path.islink(enabled_path) and filesystem.realpath(
               enabled_path) == vhost.filep:
                # Already in shape
                vhost.enabled = True
                return None
            logger.error(
                "Could not symlink %s to %s, got error: %s", enabled_path,
                vhost.filep, err.strerror)
            errstring = ("Encountered error while trying to enable a " +
                         "newly created VirtualHost located at {0} by " +
                         "linking to it from {1}")
            raise errors.NotSupportedError(errstring.format(vhost.filep,
                                                            enabled_path))
        vhost.enabled = True
        logger.info("Enabling available site: %s", vhost.filep)
        self.save_notes += "Enabled site %s\n" % vhost.filep
        return None

    def enable_mod(self, mod_name: str, temp: bool = False) -> None:
        """Enables module in Apache.

        Both enables and reloads Apache so module is active.

        :param str mod_name: Name of the module to enable. (e.g. 'ssl')
        :param bool temp: Whether or not this is a temporary action.

        :raises .errors.NotSupportedError: If the filesystem layout is not
            supported.
        :raises .errors.MisconfigurationError: If a2enmod or a2dismod cannot be
            run.

        """
        avail_path = os.path.join(self.parser.root, "mods-available")
        enabled_path = os.path.join(self.parser.root, "mods-enabled")
        if not os.path.isdir(avail_path) or not os.path.isdir(enabled_path):
            raise errors.NotSupportedError(
                "Unsupported directory layout. You may try to enable mod %s "
                "and try again." % mod_name)

        deps = apache_util.get_mod_deps(mod_name)

        # Enable all dependencies
        for dep in deps:
            if (dep + "_module") not in self.parser.modules:
                self._enable_mod_debian(dep, temp)
                self.parser.add_mod(dep)
                note = "Enabled dependency of %s module - %s" % (mod_name, dep)
                if not temp:
                    self.save_notes += note + os.linesep
                logger.debug(note)

        # Enable actual module
        self._enable_mod_debian(mod_name, temp)
        self.parser.add_mod(mod_name)

        if not temp:
            self.save_notes += "Enabled %s module in Apache\n" % mod_name
        logger.info("Enabled Apache %s module", mod_name)

        # Modules can enable additional config files. Variables may be defined
        # within these new configuration sections.
        # Reload is not necessary as DUMP_RUN_CFG uses latest config.
        self.parser.update_runtime_variables()

    def _enable_mod_debian(self, mod_name: str, temp: bool) -> None:
        """Assumes mods-available, mods-enabled layout."""
        # Generate reversal command.
        # Try to be safe here... check that we can probably reverse before
        # applying enmod command
        if (self.options.dismod is None or self.options.enmod is None
                or not util.exe_exists(self.options.dismod)):
            raise errors.MisconfigurationError(
                "Unable to find a2dismod, please make sure a2enmod and "
                "a2dismod are configured correctly for certbot.")

        self.reverter.register_undo_command(temp, [self.options.dismod, "-f", mod_name])
        util.run_script([self.options.enmod, mod_name])



"""Tests for certbot._internal.auth_handler."""
import datetime
import logging
import sys
import unittest
from unittest import mock

from josepy import b64encode
import pytest

from acme import challenges
from acme import client as acme_client
from acme import errors as acme_errors
from acme import messages
from certbot import achallenges
from certbot import errors
from certbot._internal.display import obj as display_obj
from certbot.plugins import common as plugin_common
from certbot.tests import acme_util
from certbot.tests import util as test_util


class ChallengeFactoryTest(unittest.TestCase):
    # pylint: disable=protected-access

    def setUp(self):
        from certbot._internal.auth_handler import AuthHandler

        # Account is mocked...
        self.handler = AuthHandler(None, None, mock.Mock(key="mock_key"), [])

        self.authzr = acme_util.gen_authzr(
            messages.STATUS_PENDING, "test", acme_util.CHALLENGES,
            [messages.STATUS_PENDING] * 6)

    def test_all(self):
        achalls = self.handler._challenge_factory(
            self.authzr, range(0, len(acme_util.CHALLENGES)))

        assert [achall.chall for achall in achalls] == acme_util.CHALLENGES

    def test_one_http(self):
        achalls = self.handler._challenge_factory(self.authzr, [0])

        assert [achall.chall for achall in achalls] == [acme_util.HTTP01]

    def test_unrecognized(self):
        authzr = acme_util.gen_authzr(
            messages.STATUS_PENDING, "test",
            [mock.Mock(chall="chall", typ="unrecognized")],
            [messages.STATUS_PENDING])

        achalls = self.handler._challenge_factory(authzr, [0])
        assert type(achalls[0]) is achallenges.Other


class HandleAuthorizationsTest(unittest.TestCase):
    """handle_authorizations test.

    This tests everything except for all functions under _poll_challenges.

    """

    def setUp(self):
        from certbot._internal.auth_handler import AuthHandler

        self.mock_display = mock.Mock()
        self.mock_config = mock.Mock(debug_challenges=False)
        display_obj.set_display(self.mock_display)

        self.mock_auth = mock.MagicMock(name="Authenticator")

        self.mock_auth.get_chall_pref.return_value = [challenges.HTTP01]

        self.mock_auth.perform.side_effect = gen_auth_resp

        self.mock_account = mock.MagicMock()
        self.mock_net = mock.MagicMock(spec=acme_client.ClientV2)
        self.mock_net.retry_after.side_effect = acme_client.ClientV2.retry_after

        self.handler = AuthHandler(
            self.mock_auth, self.mock_net, self.mock_account, [])

        logging.disable(logging.CRITICAL)

    def tearDown(self):
        logging.disable(logging.NOTSET)

    def _test_name1_http_01_1_common(self):
        authzr = gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)
        mock_order = mock.MagicMock(authorizations=[authzr])

        self.mock_net.poll.side_effect = _gen_mock_on_poll(retry=1, wait_value=30)
        with mock.patch('certbot._internal.auth_handler.time') as mock_time:
            authzr = self.handler.handle_authorizations(mock_order, self.mock_config)

            assert self.mock_net.answer_challenge.call_count == 1

            assert self.mock_net.poll.call_count == 2  # Because there is one retry
            assert mock_time.sleep.call_count == 2
            # Retry-After header is 30 seconds, but at the time sleep is invoked, several
            # instructions are executed, and next pool is in less than 30 seconds.
            assert mock_time.sleep.call_args_list[1][0][0] <= 30
            # However, assert that we did not took the default value of 3 seconds.
            assert mock_time.sleep.call_args_list[1][0][0] > 3

            assert self.mock_auth.cleanup.call_count == 1
            # Test if list first element is http-01, use typ because it is an achall
            assert self.mock_auth.cleanup.call_args[0][0][0].typ == "http-01"

            assert len(authzr) == 1

    def test_name1_http_01_1_acme_2(self):
        self._test_name1_http_01_1_common()

    def test_name1_http_01_1_dns_1_acme_2(self):
        self.mock_net.poll.side_effect = _gen_mock_on_poll()
        self.mock_auth.get_chall_pref.return_value.append(challenges.DNS01)

        authzr = gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)
        mock_order = mock.MagicMock(authorizations=[authzr])
        authzr = self.handler.handle_authorizations(mock_order, self.mock_config)

        assert self.mock_net.answer_challenge.call_count == 1

        assert self.mock_net.poll.call_count == 1

        assert self.mock_auth.cleanup.call_count == 1
        cleaned_up_achalls = self.mock_auth.cleanup.call_args[0][0]
        assert len(cleaned_up_achalls) == 1
        assert cleaned_up_achalls[0].typ == "http-01"

        # Length of authorizations list
        assert len(authzr) == 1

    def test_name3_http_01_3_common_acme_2(self):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES),
                   gen_dom_authzr(domain="1", challs=acme_util.CHALLENGES),
                   gen_dom_authzr(domain="2", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        self.mock_net.poll.side_effect = _gen_mock_on_poll()
        authzr = self.handler.handle_authorizations(mock_order, self.mock_config)

        assert self.mock_net.answer_challenge.call_count == 3

        # Check poll call
        assert self.mock_net.poll.call_count == 3

        assert self.mock_auth.cleanup.call_count == 1

        assert len(authzr) == 3

    def test_debug_challenges(self):
        config = mock.Mock(debug_challenges=True, verbose_count=0)
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        account_key_thumbprint = b"foobarbaz"
        self.mock_account.key.thumbprint.return_value = account_key_thumbprint

        self.mock_net.poll.side_effect = _gen_mock_on_poll()

        self.handler.handle_authorizations(mock_order, config)

        assert self.mock_net.answer_challenge.call_count == 1
        assert self.mock_display.notification.call_count == 1
        assert 'Pass "-v" for more info' in \
                      self.mock_display.notification.call_args[0][0]
        assert f"http://{authzrs[0].body.identifier.value}/.well-known/acme-challenge/" + \
                         b64encode(authzrs[0].body.challenges[0].chall.token).decode() not in \
                         self.mock_display.notification.call_args[0][0]
        assert b64encode(account_key_thumbprint).decode() not in \
                      self.mock_display.notification.call_args[0][0]

    def test_debug_challenges_verbose(self):
        config = mock.Mock(debug_challenges=True, verbose_count=1)
        authzrs = [gen_dom_authzr(domain="0", challs=[acme_util.HTTP01]),
                   gen_dom_authzr(domain="1", challs=[acme_util.DNS01])]
        mock_order = mock.MagicMock(authorizations=authzrs)

        account_key_thumbprint = b"foobarbaz"
        self.mock_account.key.thumbprint.return_value = account_key_thumbprint

        self.mock_net.poll.side_effect = _gen_mock_on_poll()

        self.mock_auth.get_chall_pref.return_value = [challenges.HTTP01,
                                                      challenges.DNS01]

        self.handler.handle_authorizations(mock_order, config)

        assert self.mock_net.answer_challenge.call_count == 2
        assert self.mock_display.notification.call_count == 1
        assert 'Pass "-v" for more info' not in \
                         self.mock_display.notification.call_args[0][0]
        assert f"http://{authzrs[0].body.identifier.value}/.well-known/acme-challenge/" + \
                      b64encode(authzrs[0].body.challenges[0].chall.token).decode() in \
                      self.mock_display.notification.call_args[0][0]
        assert b64encode(account_key_thumbprint).decode() in \
                      self.mock_display.notification.call_args[0][0]
        assert f"_acme-challenge.{authzrs[1].body.identifier.value}" in \
                      self.mock_display.notification.call_args[0][0]
        assert authzrs[1].body.challenges[0].validation(self.mock_account.key) in \
                      self.mock_display.notification.call_args[0][0]

    def test_perform_failure(self):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        self.mock_auth.perform.side_effect = errors.AuthorizationError

        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)

    def test_max_retries_exceeded(self):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        # We will return STATUS_PENDING twice before returning STATUS_VALID.
        self.mock_net.poll.side_effect = _gen_mock_on_poll(retry=2)

        with pytest.raises(errors.AuthorizationError,
                           match='All authorizations were not finalized by the CA.'):
            # We retry only once, so retries will be exhausted before STATUS_VALID is returned.
            self.handler.handle_authorizations(mock_order, self.mock_config, False, 1)

    @mock.patch('certbot._internal.auth_handler.time.sleep')
    def test_deadline_exceeded(self, mock_sleep):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        orig_now = datetime.datetime.now
        state = {'time_slept': 0}

        def mock_sleep_effect(secs):
            state['time_slept'] += secs
        mock_sleep.side_effect = mock_sleep_effect

        def mock_now_effect():
            return orig_now() + datetime.timedelta(seconds=state["time_slept"])

        # We will return STATUS_PENDING and ask Certbot to sleep for 20 minutes at a time.
        interval = datetime.timedelta(minutes=20).seconds
        self.mock_net.poll.side_effect = _gen_mock_on_poll(status=messages.STATUS_PENDING,
                                                           wait_value=interval)

        with pytest.raises(errors.AuthorizationError,
                           match='All authorizations were not finalized by the CA.'):
            with mock.patch('certbot._internal.auth_handler.datetime.datetime') as mock_dt:
                mock_dt.now.side_effect = mock_now_effect
                # Polling will only proceed for 30 minutes at most, so the second 20 minute sleep
                # should be truncated and the polling should be aborted.
                self.handler.handle_authorizations(mock_order, self.mock_config, False)

        assert mock_sleep.call_count == 3 # 1s, 20m and 10m sleep
        assert mock_sleep.call_args_list[0][0][0] == 1
        assert abs(mock_sleep.call_args_list[1][0][0] - (interval - 1)) <= 1
        assert abs(mock_sleep.call_args_list[2][0][0] - (interval/2 - 1)) <= 1

    def test_no_domains(self):
        mock_order = mock.MagicMock(authorizations=[])
        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)

    def test_preferred_challenge_choice_common_acme_2(self):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        self.mock_auth.get_chall_pref.return_value.append(challenges.HTTP01)

        self.handler.pref_challs.extend((challenges.HTTP01.typ,
                                         challenges.DNS01.typ,))

        self.mock_net.poll.side_effect = _gen_mock_on_poll()
        self.handler.handle_authorizations(mock_order, self.mock_config)

        assert self.mock_auth.cleanup.call_count == 1
        assert self.mock_auth.cleanup.call_args[0][0][0].typ == "http-01"

    def test_preferred_challenges_not_supported_acme_2(self):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)
        self.handler.pref_challs.append(challenges.DNS01.typ)
        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)

    def test_dns_only_challenge_not_supported(self):
        authzrs = [gen_dom_authzr(domain="0", challs=[acme_util.DNS01])]
        mock_order = mock.MagicMock(authorizations=authzrs)
        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)

    def test_perform_error(self):
        self.mock_auth.perform.side_effect = errors.AuthorizationError

        authzr = gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)
        mock_order = mock.MagicMock(authorizations=[authzr])
        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)

        assert self.mock_auth.cleanup.call_count == 1
        assert self.mock_auth.cleanup.call_args[0][0][0].typ == "http-01"

    def test_answer_error(self):
        self.mock_net.answer_challenge.side_effect = errors.AuthorizationError

        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)

        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)
        assert self.mock_auth.cleanup.call_count == 1
        assert self.mock_auth.cleanup.call_args[0][0][0].typ == "http-01"

    def test_incomplete_authzr_error(self):
        authzrs = [gen_dom_authzr(domain="0", challs=acme_util.CHALLENGES)]
        mock_order = mock.MagicMock(authorizations=authzrs)
        self.mock_net.poll.side_effect = _gen_mock_on_poll(status=messages.STATUS_INVALID)

        with test_util.patch_display_util():
            with pytest.raises(errors.AuthorizationError, match='Some challenges have failed.'):
                self.handler.handle_authorizations(mock_order, self.mock_config, False)
        assert self.mock_auth.cleanup.call_count == 1
        assert self.mock_auth.cleanup.call_args[0][0][0].typ == "http-01"

    def test_best_effort(self):
        def _conditional_mock_on_poll(authzr):
            """This mock will invalidate one authzr, and invalidate the other one"""
            valid_mock = _gen_mock_on_poll(messages.STATUS_VALID)
            invalid_mock = _gen_mock_on_poll(messages.STATUS_INVALID)

            if authzr.body.identifier.value == 'will-be-invalid':
                return invalid_mock(authzr)
            return valid_mock(authzr)

        # Two authzrs. Only one will be valid.
        authzrs = [gen_dom_authzr(domain="will-be-valid", challs=acme_util.CHALLENGES),
                   gen_dom_authzr(domain="will-be-invalid", challs=acme_util.CHALLENGES)]
        self.mock_net.poll.side_effect = _conditional_mock_on_poll

        mock_order = mock.MagicMock(authorizations=authzrs)

        with mock.patch('certbot._internal.auth_handler.AuthHandler._report_failed_authzrs') \
            as mock_report:
            valid_authzr = self.handler.handle_authorizations(mock_order, self.mock_config, True)

        # Because best_effort=True, we did not blow up. Instead ...
        assert len(valid_authzr) == 1  # ... the valid authzr has been processed
        assert mock_report.call_count == 1  # ... the invalid authzr has been reported

        self.mock_net.poll.side_effect = _gen_mock_on_poll(status=messages.STATUS_INVALID)

        with test_util.patch_display_util():
            with pytest.raises(errors.AuthorizationError, match='All challenges have failed.'):
                # Despite best_effort=True, process will fail because no authzr is valid.
                self.handler.handle_authorizations(mock_order, self.mock_config, True)

    def test_validated_challenge_not_rerun(self):
        # With a pending challenge that is not supported by the plugin, we
        # expect an exception to be raised.
        authzr = acme_util.gen_authzr(
                messages.STATUS_PENDING, "0",
                [acme_util.DNS01],
                [messages.STATUS_PENDING])
        mock_order = mock.MagicMock(authorizations=[authzr])
        with pytest.raises(errors.AuthorizationError):
            self.handler.handle_authorizations(mock_order, self.mock_config)

        # With a validated challenge that is not supported by the plugin, we
        # expect the challenge to not be solved again and
        # handle_authorizations() to succeed.
        authzr = acme_util.gen_authzr(
                messages.STATUS_VALID, "0",
                [acme_util.DNS01],
                [messages.STATUS_VALID])
        mock_order = mock.MagicMock(authorizations=[authzr])
        self.handler.handle_authorizations(mock_order, self.mock_config)

    def test_valid_authzrs_deactivated(self):
        """When we deactivate valid authzrs in an orderr, we expect them to become deactivated
        and to receive a list of deactivated authzrs in return."""
        def _mock_deactivate(authzr):
            if authzr.body.status == messages.STATUS_VALID:
                if authzr.body.identifier.value == "is_valid_but_will_fail":
                    raise acme_errors.Error("Mock deactivation ACME error")
                authzb = authzr.body.update(status=messages.STATUS_DEACTIVATED)
                authzr = messages.AuthorizationResource(body=authzb)
            else: # pragma: no cover
                raise errors.Error("Can't deactivate non-valid authz")
            return authzr

        to_deactivate = [("is_valid", messages.STATUS_VALID),
                         ("is_pending", messages.STATUS_PENDING),
                         ("is_valid_but_will_fail", messages.STATUS_VALID)]

        to_deactivate = [acme_util.gen_authzr(a[1], a[0], [acme_util.HTTP01],
                         [a[1]]) for a in to_deactivate]
        orderr = mock.MagicMock(authorizations=to_deactivate)

        self.mock_net.deactivate_authorization.side_effect = _mock_deactivate

        authzrs, failed = self.handler.deactivate_valid_authorizations(orderr)

        assert self.mock_net.deactivate_authorization.call_count == 2
        assert len(authzrs) == 1
        assert len(failed) == 1
        assert authzrs[0].body.identifier.value == "is_valid"
        assert authzrs[0].body.status == messages.STATUS_DEACTIVATED
        assert failed[0].body.identifier.value == "is_valid_but_will_fail"
        assert failed[0].body.status == messages.STATUS_VALID


def _gen_mock_on_poll(status=messages.STATUS_VALID, retry=0, wait_value=1):
    state = {'count': retry}

    def _mock(authzr):
        state['count'] = state['count'] - 1
        effective_status = status if state['count'] < 0 else messages.STATUS_PENDING
        updated_azr = acme_util.gen_authzr(
            effective_status,
            authzr.body.identifier.value,
            [challb.chall for challb in authzr.body.challenges],
            [effective_status] * len(authzr.body.challenges))
        return updated_azr, mock.MagicMock(headers={'Retry-After': str(wait_value)})
    return _mock


class ChallbToAchallTest(unittest.TestCase):
    """Tests for certbot._internal.auth_handler.challb_to_achall."""

    def _call(self, challb):
        from certbot._internal.auth_handler import challb_to_achall
        ident = messages.Identifier(typ=messages.IDENTIFIER_FQDN, value="domain")
        return challb_to_achall(challb, "account_key", ident)

    def test_it(self):
        assert self._call(acme_util.HTTP01_P) == \
            achallenges.KeyAuthorizationAnnotatedChallenge(
                challb=acme_util.HTTP01_P, account_key="account_key",
                identifier=messages.Identifier(typ=messages.IDENTIFIER_FQDN, value="domain"))


class GenChallengePathTest(unittest.TestCase):
    """Tests for certbot._internal.auth_handler.gen_challenge_path.

    """
    def setUp(self):
        logging.disable(logging.FATAL)

    def tearDown(self):
        logging.disable(logging.NOTSET)

    @classmethod
    def _call(cls, challbs, preferences):
        from certbot._internal.auth_handler import gen_challenge_path
        return gen_challenge_path(challbs, preferences)

    def test_common_case(self):
        """Given DNS01 and HTTP01 with appropriate combos."""
        challbs = (acme_util.DNS01_P, acme_util.HTTP01_P)
        prefs = [challenges.DNS01, challenges.HTTP01]

        assert self._call(challbs, prefs) == (0,)
        assert self._call(challbs[::-1], prefs) == (1,)

    def test_not_supported(self):
        challbs = (acme_util.DNS01_P,)
        prefs = [challenges.HTTP01]

        # smart path fails because no challs in prefs satisfies combos
        with pytest.raises(errors.AuthorizationError):
            self._call(challbs, prefs)


class ReportFailedAuthzrsTest(unittest.TestCase):
    """Tests for certbot._internal.auth_handler.AuthHandler._report_failed_authzrs."""
    # pylint: disable=protected-access


    def setUp(self):
        from certbot._internal.auth_handler import AuthHandler

        self.mock_auth = mock.MagicMock(spec=plugin_common.Plugin, name="buzz")
        self.mock_auth.name = "buzz"
        self.mock_auth.auth_hint.return_value = "the buzz hint"
        self.handler = AuthHandler(self.mock_auth, mock.MagicMock(), mock.MagicMock(), [])

        kwargs = {
            "chall": acme_util.HTTP01,
            "uri": "uri",
            "status": messages.STATUS_INVALID,
            "error": messages.Error.with_code("tls", detail="detail"),
        }

        # Prevent future regressions if the error type changes
        assert kwargs["error"].description is not None

        http_01 = messages.ChallengeBody(**kwargs)

        kwargs["chall"] = acme_util.HTTP01
        http_01 = messages.ChallengeBody(**kwargs)

        self.authzr1 = mock.MagicMock()
        self.authzr1.body.identifier.value = 'example.com'
        self.authzr1.body.challenges = [http_01, http_01]

        kwargs["error"] = messages.Error.with_code("dnssec", detail="detail")
        http_01_diff = messages.ChallengeBody(**kwargs)

        self.authzr2 = mock.MagicMock()
        self.authzr2.body.identifier.value = 'foo.bar'
        self.authzr2.body.challenges = [http_01_diff]

    @mock.patch('certbot._internal.auth_handler.display_util.notify')
    def test_same_error_and_domain(self, mock_notify):
        self.handler._report_failed_authzrs([self.authzr1])
        mock_notify.assert_called_with(
            '\n'
            'Certbot failed to authenticate some domains (authenticator: buzz). '
            'The Certificate Authority reported these problems:\n'
            '  Identifier: example.com\n'
            '  Type:   tls\n'
            '  Detail: detail\n'
            '\n'
            '  Identifier: example.com\n'
            '  Type:   tls\n'
            '  Detail: detail\n'
            '\nHint: the buzz hint\n'
        )

    @mock.patch('certbot._internal.auth_handler.display_util.notify')
    def test_different_errors_and_domains(self, mock_notify):
        self.mock_auth.name = "quux"
        self.mock_auth.auth_hint.return_value = "quuuuuux"
        self.handler._report_failed_authzrs([self.authzr1, self.authzr2])
        mock_notify.assert_called_with(
            '\n'
            'Certbot failed to authenticate some domains (authenticator: quux). '
            'The Certificate Authority reported these problems:\n'
            '  Identifier: foo.bar\n'
            '  Type:   dnssec\n'
            '  Detail: detail\n'
            '\n'
            '  Identifier: example.com\n'
            '  Type:   tls\n'
            '  Detail: detail\n'
            '\n'
            '  Identifier: example.com\n'
            '  Type:   tls\n'
            '  Detail: detail\n'
            '\nHint: quuuuuux\n'
        )

    @mock.patch('certbot._internal.auth_handler.display_util.notify')
    def test_non_subclassed_authenticator(self, mock_notify):
        """If authenticator not derived from common.Plugin, we shouldn't call .auth_hint"""
        from certbot._internal.auth_handler import AuthHandler

        self.mock_auth = mock.MagicMock(name="quuz")
        self.mock_auth.name = "quuz"
        self.mock_auth.auth_hint.side_effect = Exception
        self.handler = AuthHandler(self.mock_auth, mock.MagicMock(), mock.MagicMock(), [])
        self.handler._report_failed_authzrs([self.authzr1])
        assert mock_notify.call_count == 1


def gen_auth_resp(chall_list):
    """Generate a dummy authorization response."""
    return ["%s%s" % (chall.__class__.__name__, chall.identifier.value)
            for chall in chall_list]


def gen_dom_authzr(domain, challs):
    """Generates new authzr for domains."""
    return acme_util.gen_authzr(
        messages.STATUS_PENDING, domain, challs,
        [messages.STATUS_PENDING] * len(challs))


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Certbot main entry point."""
# pylint: disable=too-many-lines

from contextlib import contextmanager
import copy
import functools
import logging.handlers
import sys
from typing import cast
from typing import Generator
from typing import IO
from typing import Iterable
from typing import Optional
from typing import TypeVar
from typing import Union

import configobj
from cryptography import x509
import josepy as jose
from josepy import b64

from acme import client as acme_client
from acme import errors as acme_errors
from acme import messages as acme_messages
import certbot
from certbot import configuration
from certbot import crypto_util
from certbot import errors
from certbot import interfaces
from certbot import util
from certbot._internal import account
from certbot._internal import cert_manager
from certbot._internal import cli
from certbot._internal import client
from certbot._internal import constants
from certbot._internal import eff
from certbot._internal import hooks
from certbot._internal import log
from certbot._internal import renewal
from certbot._internal import san
from certbot._internal import snap_config
from certbot._internal import storage
from certbot._internal import updater
from certbot._internal.display import obj as display_obj
from certbot._internal.display import util as internal_display_util
from certbot._internal.plugins import disco as plugins_disco
from certbot._internal.plugins import selection as plug_sel
from certbot.compat import filesystem
from certbot.compat import misc
from certbot.compat import os
from certbot.display import ops as display_ops
from certbot.display import util as display_util
from certbot.plugins import enhancements

USER_CANCELLED = ("User chose to cancel the operation and may "
                  "reinvoke the client.")


logger = logging.getLogger(__name__)


def _suggest_donation_if_appropriate(config: configuration.NamespaceConfig) -> None:
    """Potentially suggest a donation to support Certbot.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :returns: `None`
    :rtype: None

    """
    # don't prompt for donation if:
    # - renewing
    # - using the staging server (--staging or --dry-run)
    # - running with --quiet (display fd won't be available during atexit calls #8995)
    assert config.verb != "renew"
    if config.staging or config.quiet:
        return
    util.atexit_register(
        display_util.notification,
        "If you like Certbot, please consider supporting our work by:\n"
        " * Donating to ISRG / Let's Encrypt:   https://letsencrypt.org/donate\n"
        " * Donating to EFF:                    https://eff.org/donate-le",
        pause=False
    )


def _get_and_save_cert(le_client: client.Client, config: configuration.NamespaceConfig,
                       sans: Optional[list[san.SAN]] = None, certname: Optional[str] = None,
                       lineage: Optional[storage.RenewableCert] = None
                       ) -> Optional[storage.RenewableCert]:
    """Authenticate and enroll certificate.

    This method finds the relevant lineage, figures out what to do with it,
    then performs that action. Includes calls to hooks, various reports,
    checks, and requests for user input.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param sans: List of domain names and/or IP addresses for which to get a certificate.
        Defaults to `None`
    :type sans: `list` of `san.SAN`

    :param certname: Name of new certificate. Defaults to `None`
    :type certname: str

    :param lineage: Certificate lineage object. Defaults to `None`
    :type lineage: storage.RenewableCert

    :returns: the issued certificate or `None` if doing a dry run
    :rtype: storage.RenewableCert or None

    :raises errors.Error: if certificate could not be obtained

    """
    hooks.pre_hook(config)
    renewed_sans: list[san.SAN] = []

    try:
        if lineage is not None:
            # Renewal, where we already know the specific lineage we're
            # interested in
            display_util.notify(
                "{action} for {identifiers}".format(
                    action="Simulating renewal of an existing certificate"
                    if config.dry_run else "Renewing an existing certificate",
                    identifiers=internal_display_util.summarize_sans(sans or lineage.sans())
                )
            )
            renewal.renew_cert(config, sans, le_client, lineage)
        else:
            # TREAT AS NEW REQUEST
            if sans is None:
                raise errors.Error("Domain list cannot be none if the lineage is not set.")
            display_util.notify(
                "{action} for {identifiers}".format(
                    action="Simulating a certificate request" if config.dry_run else
                           "Requesting a certificate",
                    identifiers=internal_display_util.summarize_sans(sans)
                )
            )
            lineage = le_client.obtain_and_enroll_certificate(sans, certname)
            if lineage is not None:
                hooks.deploy_hook(config, lineage.sans(), lineage.live_dir)
                renewed_sans.extend(sans)
    finally:
        hooks.post_hook(config, renewed_sans)

    return lineage


def _handle_unexpected_key_type_migration(config: configuration.NamespaceConfig,
                                          cert: storage.RenewableCert) -> bool:
    """
    This function ensures that the user will not implicitly migrate an existing key
    from one type to another in the situation where a certificate for that lineage
    already exist and they have not provided explicitly --key-type and --cert-name.
    :param config: Current configuration provided by the client
    :param cert: Matching certificate that could be renewed
    :returns: Whether a key type migration is going ahead.
    :rtype: `bool`
    """
    new_key_type = config.key_type.upper()
    cur_key_type = cert.private_key_type.upper()

    if new_key_type == cur_key_type:
        return False

    # If both --key-type and --cert-name are provided, we consider the user's intent to
    # be unambiguous: to change the key type of this lineage.
    is_confirmed_via_cli = config.set_by_user("key_type") and config.set_by_user("certname")

    # Failing that, we interactively prompt the user to confirm the change.
    if is_confirmed_via_cli or display_util.yesno(
        f'An {cur_key_type} certificate named {cert.lineagename} already exists. Do you want to '
        f'update its key type to {new_key_type}?',
        yes_label='Update key type', no_label='Keep existing key type',
        default=False, force_interactive=False,
    ):
        return True

    # If --key-type was set on the CLI but the user did not confirm the key type change using
    # one of the two above methods, their intent is ambiguous. Error out.
    if config.set_by_user("key_type"):
        raise errors.Error(
            'Are you trying to change the key type of the certificate named '
            f'{cert.lineagename} from {cur_key_type} to {new_key_type}? Please provide '
            'both --cert-name and --key-type on the command line to confirm the change '
            'you are trying to make.'
        )

    # The mismatch between the lineage's key type and config.key_type is caused by Certbot's
    # default value. The user is not asking for a key change: keep the key type of the existing
    # lineage.
    config.key_type = cur_key_type.lower()
    return False


def _handle_subset_cert_request(config: configuration.NamespaceConfig,
                                sans: Iterable[san.SAN],
                                cert: storage.RenewableCert
                                ) -> tuple[str, Optional[storage.RenewableCert]]:
    """Figure out what to do if a previous cert had a subset of the names now requested

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param sans: List of domain names and/or IP addresses
    :type sans: `list` of `san.SAN`

    :param cert: Certificate object
    :type cert: storage.RenewableCert

    :returns: Tuple of (str action, cert_or_None) as per _find_lineage_for_sans_and_certname
              action can be: "newcert" | "renew" | "reinstall"
    :rtype: `tuple` of `str`

    """
    _handle_unexpected_key_type_migration(config, cert)

    question = (
        "You have an existing certificate that contains a portion of "
        "the domains you requested (ref: {0}){br}{br}It contains these "
        "names: {1}{br}{br}You requested these names for the new "
        "certificate: {2}.{br}{br}Do you want to expand and replace this existing "
        "certificate with the new certificate?"
    ).format(cert.configfile.filename,
             san.display(cert.sans()),
             san.display(sans),
             br=os.linesep)
    if config.expand or config.renew_by_default or display_util.yesno(
        question, "Expand", "Cancel", cli_flag="--expand", force_interactive=True):
        return "renew", cert
    display_util.notify(
        "To obtain a new certificate that contains these names without "
        "replacing your existing certificate for {0}, you must use the "
        "--duplicate option.{br}{br}"
        "For example:{br}{br}{1} --duplicate {2}".format(
            san.display(cert.sans()),
            cli.cli_command, " ".join(sys.argv[1:]),
            br=os.linesep
        ))
    raise errors.Error(USER_CANCELLED)


def _handle_identical_cert_request(config: configuration.NamespaceConfig,
                                   lineage: storage.RenewableCert,
                                   ) -> tuple[str, Optional[storage.RenewableCert]]:
    """Figure out what to do if a lineage has the same names as a previously obtained one

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param lineage: Certificate lineage object
    :type lineage: storage.RenewableCert

    :returns: Tuple of (str action, cert_or_None) as per _find_lineage_for_sans_and_certname
              action can be: "newcert" | "renew" | "reinstall"
    :rtype: `tuple` of `str`

    """
    is_key_type_changing = _handle_unexpected_key_type_migration(config, lineage)

    if not lineage.ensure_deployed():
        return "reinstall", lineage
    if is_key_type_changing:
        return "renew", lineage

    ari_clients = renewal.AriClientPool(config)

    if renewal.should_renew(config, lineage, ari_clients):
        return "renew", lineage

    if config.reinstall:
        # Set with --reinstall, force an identical certificate to be
        # reinstalled without further prompting.
        return "reinstall", lineage
    question = (
        "You have an existing certificate that has exactly the same "
        "domains or certificate name you requested and isn't close to expiry."
        "{br}(ref: {0}){br}{br}What would you like to do?"
    ).format(lineage.configfile.filename, br=os.linesep)

    if config.verb == "run":
        keep_opt = "Attempt to reinstall this existing certificate"
    else:
        assert config.verb == "certonly", "Unexpected Certbot subcommand"
        keep_opt = "Keep the existing certificate for now"
    choices = [keep_opt,
               "Renew & replace the certificate (may be subject to CA rate limits)"]

    response = display_util.menu(question, choices,
                                    default=0, force_interactive=True)
    if response[0] == display_util.CANCEL:
        # TODO: Add notification related to command-line options for
        #       skipping the menu for this case.
        raise errors.Error(
            "Operation canceled. You may re-run the client.")
    if response[1] == 0:
        return "reinstall", lineage
    elif response[1] == 1:
        return "renew", lineage
    raise AssertionError('This is impossible')


def _find_lineage_for_sans(config: configuration.NamespaceConfig, sans: list[san.SAN]
                           ) -> tuple[Optional[str], Optional[storage.RenewableCert]]:
    """Determine whether there are duplicated names and how to handle
    them (renew, reinstall, newcert, or raising an error to stop
    the client run if the user chooses to cancel the operation when
    prompted).

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param sans: List of domain names and/or IP addresses
    :type sans: `list` of `san.SAN`

    :returns: Two-element tuple containing desired new-certificate behavior as
              a string token ("reinstall", "renew", or "newcert"), plus either
              a RenewableCert instance or `None` if renewal shouldn't occur.
    :rtype: `tuple` of `str` and :class:`storage.RenewableCert` or `None`

    :raises errors.Error: If the user would like to rerun the client again.

    """
    # Considering the possibility that the requested certificate is
    # related to an existing certificate.  (config.duplicate, which
    # is set with --duplicate, skips all of this logic and forces any
    # kind of certificate to be obtained with renewal = False.)
    if config.duplicate:
        return "newcert", None
    # TODO: Also address superset case
    ident_names_cert, subset_names_cert = cert_manager.find_duplicative_certs(config, sans)
    # XXX ^ schoen is not sure whether that correctly reads the systemwide
    # configuration file.
    if ident_names_cert is None and subset_names_cert is None:
        return "newcert", None

    if ident_names_cert is not None:
        return _handle_identical_cert_request(config, ident_names_cert)
    elif subset_names_cert is not None:
        return _handle_subset_cert_request(config, sans, subset_names_cert)
    return None, None


def _find_cert(config: configuration.NamespaceConfig, sans: list[san.SAN], certname: str
               ) -> tuple[bool, Optional[storage.RenewableCert]]:
    """Finds an existing certificate object given domains and/or a certificate name.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param sans: List of domain names and/or IP addresses
    :type sans: `list` of `san.SAN`

    :param certname: Name of certificate
    :type certname: str

    :returns: Two-element tuple of a boolean that indicates if this function should be
              followed by a call to fetch a certificate from the server, and either a
              RenewableCert instance or None.
    :rtype: `tuple` of `bool` and :class:`storage.RenewableCert` or `None`

    """
    action, lineage = _find_lineage_for_sans_and_certname(config, sans, certname)
    if action == "reinstall":
        logger.info("Keeping the existing certificate")
    return (action != "reinstall"), lineage


def _find_lineage_for_sans_and_certname(
        config: configuration.NamespaceConfig, sans: list[san.SAN],
        certname: str) -> tuple[Optional[str], Optional[storage.RenewableCert]]:
    """Find appropriate lineage based on given domains and/or certname.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param sans: List of domain names and/or IP addresses
    :type sans: `list` of `san.SAN`

    :param certname: Name of certificate
    :type certname: str

    :returns: Two-element tuple containing desired new-certificate behavior as
              a string token ("reinstall", "renew", or "newcert"), plus either
              a RenewableCert instance or None if renewal should not occur.

    :rtype: `tuple` of `str` and :class:`storage.RenewableCert` or `None`

    :raises errors.Error: If the user would like to rerun the client again.

    """
    if not certname:
        return _find_lineage_for_sans(config, sans)
    lineage = cert_manager.lineage_for_certname(config, certname)
    if lineage:
        if sans:
            computed_domains = cert_manager.sans_for_certname(config, certname)
            if computed_domains and set(computed_domains) != set(sans):
                _handle_unexpected_key_type_migration(config, lineage)
                _ask_user_to_confirm_new_sans(config, sans, certname,
                                              lineage.sans())  # raises if no
                return "renew", lineage
        # unnecessarily specified domains or no domains specified
        return _handle_identical_cert_request(config, lineage)
    elif sans:
        return "newcert", None
    raise errors.ConfigurationError("No certificate with name {0} found. "
                                    "Use -d to specify domains, or run certbot certificates to see "
                                    "possible certificate names.".format(certname))


T = TypeVar("T")


def _get_added_removed(after: Iterable[T], before: Iterable[T]) -> tuple[list[T], list[T]]:
    """Get lists of items removed from `before`
    and a lists of items added to `after`
    """
    added = list(set(after) - set(before))
    removed = list(set(before) - set(after))
    added.sort()
    removed.sort()
    return added, removed


def _format_list(character: str, strings: Iterable[str]) -> str:
    """Format list with given character
    """
    if not strings:
        formatted = "{br}(None)"
    else:
        formatted = "{br}{ch} " + "{br}{ch} ".join(strings)
    return formatted.format(
        ch=character,
        br=os.linesep
    )


def _ask_user_to_confirm_new_sans(config: configuration.NamespaceConfig,
                                  new_sans: Iterable[san.SAN],
                                  certname: str,
                                  old_sans: Iterable[san.SAN]) -> None:
    """Ask user to confirm update cert certname to contain new_sans.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param new_sans: List of new domain names and/or IP addresses
    :type new_sans: `list` of `san.SAN`

    :param certname: Name of certificate
    :type certname: str

    :param old_sans: List of old domain names and/or IP addresses
    :type old_sans: `list` of `san.SAN`

    :returns: None
    :rtype: None

    :raises errors.ConfigurationError: if cert name and domains mismatch

    """
    if config.renew_with_new_domains:
        return

    added, removed = _get_added_removed(map(str, new_sans), map(str, old_sans))

    msg = ("You are updating certificate {0} to include new domain(s): {1}{br}{br}"
           "You are also removing previously included domain(s): {2}{br}{br}"
           "Did you intend to make this change?".format(
               certname,
               _format_list("+", added),
               _format_list("-", removed),
               br=os.linesep))
    if not display_util.yesno(msg, "Update certificate", "Cancel", default=True):
        raise errors.ConfigurationError("Specified mismatched certificate name and domains.")


def _find_sans_or_certname(config: configuration.NamespaceConfig,
                           installer: Optional[interfaces.Installer],
                           question: Optional[str] = None) -> tuple[list[san.SAN], str]:
    """Retrieve domains and certname from config or user input.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param installer: Installer object
    :type installer: interfaces.Installer

    :param `str` question: Overriding default question to ask the user if asked
        to choose from domain names.

    :returns: Two-part tuple of domains and certname
    :rtype: `tuple` of list of `str` and `str`

    :raises errors.Error: Usage message, if parameters are not used correctly

    """
    domains: list[san.SAN] = []
    ip_addresses: list[san.SAN] = []
    certname = config.certname

    # first, try to get domains from the config
    if config.domains:
        domains = config.domains
    if config.ip_addresses:
        ip_addresses = config.ip_addresses

    sans: Optional[list[san.SAN]] = domains + ip_addresses

    # if we can't do that but we have a certname, get the sans
    # by loading the latest certificate with that certname
    if certname and not sans:
        sans = cert_manager.sans_for_certname(config, certname)

    # that certname might not have existed, or there was a problem.
    # try to get domains from the user.
    if not sans:
        sans = san.guess(display_ops.choose_names(installer, question))

    if not sans:
        raise errors.Error("Please specify --domains, --ip-address, or --installer that "
                           "will help in domain names autodiscovery, or "
                           "--cert-name for an existing certificate name.")

    return sans, certname


def _report_next_steps(config: configuration.NamespaceConfig, installer_err: Optional[errors.Error],
                       lineage: Optional[storage.RenewableCert],
                       new_or_renewed_cert: bool = True) -> None:
    """Displays post-run/certonly advice to the user about renewal and installation.

    The output varies by runtime configuration and any errors encountered during installation.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param installer_err: The installer/enhancement error encountered, if any.
    :type error: Optional[errors.Error]

    :param lineage: The resulting certificate lineage from the issuance, if any.
    :type lineage: Optional[storage.RenewableCert]

    :param bool new_or_renewed_cert: Whether the verb execution resulted in a certificate
                                     being saved (created or renewed).

    """
    steps: list[str] = []

    # If the installation or enhancement raised an error, show advice on trying again
    if installer_err:
        # Special case where either --nginx or --apache were used, causing us to
        # run the "installer" (i.e. reloading the nginx/apache config)
        if config.verb == 'certonly':
            steps.append(
                "The certificate was saved, but was not successfully loaded by the installer "
                f"({config.installer}) due to the installer failing to reload. "
                f"After fixing the error shown below, try reloading {config.installer} manually."
            )
        else:
            steps.append(
                "The certificate was saved, but could not be installed (installer: "
                f"{config.installer}). After fixing the error shown below, try installing it again "
                f"by running:\n  {cli.cli_command} install --cert-name "
                f"{_cert_name_from_config_or_lineage(config, lineage)}"
            )

    # If a certificate was obtained or renewed, show applicable renewal advice
    if new_or_renewed_cert:
        if config.csr:
            steps.append(
                "Certificates created using --csr will not be renewed automatically by Certbot. "
                "You will need to renew the certificate before it expires, by running the same "
                "Certbot command again.")
        elif _is_interactive_only_auth(config):
            steps.append(
                "This certificate will not be renewed automatically. Autorenewal of "
                "--manual certificates requires the use of an authentication hook script "
                "(--manual-auth-hook) but one was not provided. To renew this certificate, repeat "
                f"this same {cli.cli_command} command before the certificate's expiry date."
            )
        elif not config.preconfigured_renewal:
            steps.append(
                "The certificate will need to be renewed before it expires. Certbot can "
                "automatically renew the certificate in the background, but you may need "
                "to take steps to enable that functionality. "
                "See https://certbot.org/renewal-setup for instructions.")

    if not steps:
        return

    # TODO: refactor ANSI escapes during https://github.com/certbot/certbot/issues/8848
    (bold_on, nl, bold_off) = [c if sys.stdout.isatty() and not config.quiet else '' \
                               for c in (util.ANSI_SGR_BOLD, '\n', util.ANSI_SGR_RESET)]
    print(bold_on, end=nl)
    display_util.notify("NEXT STEPS:")
    print(bold_off, end='')

    for step in steps:
        display_util.notify(f"- {step}")

    # If there was an installer error, segregate the error output with a trailing newline
    if installer_err:
        print()


def _report_new_cert(config: configuration.NamespaceConfig, cert_path: Optional[str],
                     fullchain_path: Optional[str], key_path: Optional[str] = None) -> None:
    """Reports the creation of a new certificate to the user.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param cert_path: path to certificate
    :type cert_path: str

    :param fullchain_path: path to full chain
    :type fullchain_path: str

    :param key_path: path to private key, if available
    :type key_path: str

    :returns: `None`
    :rtype: None

    """
    if config.dry_run:
        display_util.notify("The dry run was successful.")
        return

    assert cert_path and fullchain_path, "No certificates saved to report."

    renewal_msg = ""
    if config.preconfigured_renewal and not _is_interactive_only_auth(config):
        renewal_msg = ("\nCertbot has set up a scheduled task to automatically renew this "
                       "certificate in the background.")

    display_util.notify(
        ("\nSuccessfully received certificate.\n"
        "Certificate is saved at: {cert_path}\n{key_msg}"
        "This certificate expires on {expiry}.\n"
        "These files will be updated when the certificate renews.{renewal_msg}{nl}").format(
            cert_path=fullchain_path,
            expiry=crypto_util.notAfter(cert_path).date(),
            key_msg="Key is saved at:         {}\n".format(key_path) if key_path else "",
            renewal_msg=renewal_msg,
            nl="\n" if config.verb == "run" else "" # Normalize spacing across verbs
        )
    )


def _is_interactive_only_auth(config: configuration.NamespaceConfig) -> bool:
    """ Whether the current authenticator params only support interactive renewal.
    """
    # --manual without --manual-auth-hook can never autorenew
    if config.authenticator == "manual" and config.manual_auth_hook is None:
        return True

    return False


def _csr_report_new_cert(config: configuration.NamespaceConfig, cert_path: Optional[str],
                         chain_path: Optional[str], fullchain_path: Optional[str]) -> None:
    """ --csr variant of _report_new_cert.

    Until --csr is overhauled (#8332) this is transitional function to report the creation
    of a new certificate using --csr.
    TODO: remove this function and just call _report_new_cert when --csr is overhauled.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param str cert_path: path to cert.pem

    :param str chain_path: path to chain.pem

    :param str fullchain_path: path to fullchain.pem

    """
    if config.dry_run:
        display_util.notify("The dry run was successful.")
        return

    assert cert_path and fullchain_path, "No certificates saved to report."

    expiry = crypto_util.notAfter(cert_path).date()

    display_util.notify(
        ("\nSuccessfully received certificate.\n"
        "Certificate is saved at:            {cert_path}\n"
        "Intermediate CA chain is saved at:  {chain_path}\n"
        "Full certificate chain is saved at: {fullchain_path}\n"
        "This certificate expires on {expiry}.").format(
            cert_path=cert_path, chain_path=chain_path,
            fullchain_path=fullchain_path, expiry=expiry,
        )
    )


def _determine_account(config: configuration.NamespaceConfig
                       ) -> tuple[account.Account,
                                  Optional[acme_client.ClientV2]]:
    """Determine which account to use.

    If ``config.account`` is ``None``, it will be updated based on the
    user input. Same for ``config.email``.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :returns: Account and optionally ACME client API (biproduct of new
        registration).
    :rtype: tuple of :class:`certbot._internal.account.Account` and :class:`acme.client.Client`

    :raises errors.Error: If unable to register an account with ACME server

    """
    def _tos_cb(terms_of_service: str) -> None:
        if config.tos:
            return
        msg = ("Please read the Terms of Service at: {0}\n"
               "You must agree in order to register with the ACME "
               "server. Do you agree?".format(terms_of_service))
        result = display_util.yesno(msg, cli_flag="--agree-tos", force_interactive=True)
        if not result:
            raise errors.Error(
                "Registration cannot proceed without accepting "
                "Terms of Service.")

    account_storage = account.AccountFileStorage(config)
    acme: Optional[acme_client.ClientV2] = None

    if config.account is not None:
        acc = account_storage.load(config.account)
    else:
        accounts = account_storage.find_all()
        if len(accounts) > 1:
            potential_acc = display_ops.choose_account(accounts)
            if not potential_acc:
                raise errors.Error("No account has been chosen.")
            acc = potential_acc
        elif len(accounts) == 1:
            acc = accounts[0]
        else:  # no account registered yet
            if config.email is None and not config.register_unsafely_without_email:
                config.email = display_ops.get_email()
            try:
                acc, acme = client.register(
                    config, account_storage, tos_cb=_tos_cb)
                display_util.notify("Account registered.")
            except errors.MissingCommandlineFlag:
                raise
            except (errors.Error, acme_messages.Error) as err:
                logger.debug("", exc_info=True)
                if acme_messages.is_acme_error(err):
                    err_msg = internal_display_util.describe_acme_error(
                        cast(acme_messages.Error, err))
                    err_msg = f"Error returned by the ACME server: {err_msg}"
                else:
                    err_msg = str(err)
                raise errors.Error(
                    f"Unable to register an account with ACME server. {err_msg}")

    config.account = acc.id
    return acc, acme


def _delete_if_appropriate(config: configuration.NamespaceConfig) -> None:
    """Does the user want to delete their now-revoked certs? If run in non-interactive mode,
    deleting happens automatically.

    :param config: parsed command line arguments
    :type config: configuration.NamespaceConfig

    :returns: `None`
    :rtype: None

    :raises errors.Error: If anything goes wrong, including bad user input, if an overlapping
        archive dir is found for the specified lineage, etc ...
    """
    attempt_deletion = config.delete_after_revoke
    if attempt_deletion is None:
        msg = ("Would you like to delete the certificate(s) you just revoked, "
               "along with all earlier and later versions of the certificate?")
        attempt_deletion = display_util.yesno(msg, yes_label="Yes (recommended)", no_label="No",
                                              force_interactive=True, default=True)

    if not attempt_deletion:
        return

    # config.cert_path must have been set
    # config.certname may have been set
    assert config.cert_path

    if not config.certname:
        config.certname = cert_manager.cert_path_to_lineage(config)

    # don't delete if the archive_dir is used by some other lineage
    archive_dir = storage.full_archive_path(
            configobj.ConfigObj(
                storage.renewal_file_for_certname(config, config.certname),
                encoding='utf-8', default_encoding='utf-8'),
            config, config.certname)
    try:
        cert_manager.match_and_check_overlaps(config, [lambda x: archive_dir],
                                              lambda x: x.archive_dir, lambda x: x.lineagename)
    except errors.OverlappingMatchFound:
        logger.warning("Not deleting revoked certificates due to overlapping archive dirs. "
                       "More than one certificate is using %s", archive_dir)
        return
    except Exception as e:
        msg = ('config.default_archive_dir: {0}, config.live_dir: {1}, archive_dir: {2},'
               'original exception: {3}')
        msg = msg.format(config.default_archive_dir, config.live_dir, archive_dir, e)
        raise errors.Error(msg)

    cert_manager.delete(config)


def _init_le_client(config: configuration.NamespaceConfig,
                    authenticator: Optional[interfaces.Authenticator],
                    installer: Optional[interfaces.Installer]) -> client.Client:
    """Initialize Let's Encrypt Client

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param authenticator: Acme authentication handler
    :type authenticator: Optional[interfaces.Authenticator]
    :param installer: Installer object
    :type installer: interfaces.Installer

    :returns: client: Client object
    :rtype: client.Client

    """
    acc: Optional[account.Account]
    if authenticator is not None:
        # if authenticator was given, then we will need account...
        acc, acme = _determine_account(config)
        logger.debug("Picked account: %r", acc)
    else:
        acc, acme = None, None

    return client.Client(config, acc, authenticator, installer, acme=acme)


def unregister(config: configuration.NamespaceConfig,
               unused_plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Deactivate account on server

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None` or a string indicating an error
    :rtype: None or str

    """
    account_storage = account.AccountFileStorage(config)
    accounts = account_storage.find_all()

    if not accounts:
        return f"Could not find existing account for server {config.server}."
    prompt = ("Are you sure you would like to irrevocably deactivate "
              "your account?")
    wants_deactivate = display_util.yesno(prompt, yes_label='Deactivate', no_label='Abort',
                                          default=True)

    if not wants_deactivate:
        return "Deactivation aborted."

    acc, acme = _determine_account(config)
    cb_client = client.Client(config, acc, None, None, acme=acme)

    if not cb_client.acme:
        raise errors.Error("ACME client is not set.")

    # delete on boulder
    cb_client.acme.deactivate_registration(acc.regr)
    account_files = account.AccountFileStorage(config)
    # delete local account files
    account_files.delete(config.account)

    display_util.notify("Account deactivated.")
    return None


def register(config: configuration.NamespaceConfig,
             unused_plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Create accounts on the server.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None` or a string indicating an error
    :rtype: None or str

    """
    # Portion of _determine_account logic to see whether accounts already
    # exist or not.
    account_storage = account.AccountFileStorage(config)
    accounts = account_storage.find_all()

    if accounts:
        # TODO: add a flag to register a duplicate account (this will
        #       also require extending _determine_account's behavior
        #       or else extracting the registration code from there)
        return ("There is an existing account; registration of a "
                "duplicate account with this command is currently "
                "unsupported.")
    # _determine_account will register an account
    _determine_account(config)
    return None


def update_account(config: configuration.NamespaceConfig,
                   unused_plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Modify accounts on the server.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None` or a string indicating an error
    :rtype: None or str

    """
    # Portion of _determine_account logic to see whether accounts already
    # exist or not.
    account_storage = account.AccountFileStorage(config)
    accounts = account_storage.find_all()

    if not accounts:
        return f"Could not find an existing account for server {config.server}."
    if config.email is None and not config.register_unsafely_without_email:
        config.email = display_ops.get_email()

    acc, acme = _determine_account(config)
    cb_client = client.Client(config, acc, None, None, acme=acme)

    if not cb_client.acme:
        raise errors.Error("ACME client is not set.")

    # Empty list of contacts in case the user is removing all emails
    acc_contacts: Iterable[str] = ()
    if config.email:
        acc_contacts = ['mailto:' + email for email in config.email.split(',')]
    # We rely on an exception to interrupt this process if it didn't work.
    prev_regr_uri = acc.regr.uri
    acc.regr = cb_client.acme.update_registration(acc.regr.update(
        body=acc.regr.body.update(contact=acc_contacts)))
    # A v1 account being used as a v2 account will result in changing the uri to
    # the v2 uri. Since it's the same object on disk, put it back to the v1 uri
    # so that we can also continue to use the account object with acmev1.
    acc.regr = acc.regr.update(uri=prev_regr_uri)
    account_storage.update_regr(acc)

    if not config.email:
        display_util.notify("Any contact information associated "
                            "with this account has been removed.")
    else:
        eff.prepare_subscription(config, acc)
        display_util.notify("Your e-mail address was updated to {0}.".format(config.email))

    return None


def show_account(config: configuration.NamespaceConfig,
                   unused_plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Fetch account info from the ACME server and show it to the user.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None` or a string indicating an error
    :rtype: None or str

    """
    # Portion of _determine_account logic to see whether accounts already
    # exist or not.
    account_storage = account.AccountFileStorage(config)
    accounts = account_storage.find_all()

    if not accounts:
        return f"Could not find an existing account for server {config.server}."

    acc, acme = _determine_account(config)
    cb_client = client.Client(config, acc, None, None, acme=acme)

    if not cb_client.acme:
        raise errors.Error("ACME client is not set.")

    regr = cb_client.acme.query_registration(acc.regr)
    output = [f"Account details for server {config.server}:",
              f"  Account URL: {regr.uri}"]

    thumbprint = b64.b64encode(acc.key.thumbprint()).decode()
    output.append(f"  Account Thumbprint: {thumbprint}")

    emails = []

    for contact in regr.body.contact:
        if contact.startswith('mailto:'):
            emails.append(contact[7:])

    output.append("  Email contact{}: {}".format(
                            "s" if len(emails) > 1 else "",
                            ", ".join(emails) if len(emails) > 0 else "none"))

    display_util.notify("\n".join(output))

    return None


def _cert_name_from_config_or_lineage(config: configuration.NamespaceConfig,
                                      lineage: Optional[storage.RenewableCert]) -> Optional[str]:
    if lineage:
        return lineage.lineagename
    elif config.certname:
        return config.certname
    try:
        cert_name = cert_manager.cert_path_to_lineage(config)
        return cert_name
    except errors.Error:
        pass

    return None


def _install_cert(config: configuration.NamespaceConfig, le_client: client.Client,
                  sans: list[san.SAN], lineage: Optional[storage.RenewableCert] = None) -> None:
    """Install a cert

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param le_client: Client object
    :type le_client: client.Client

    :param sans: List of domains
    :type sans: `list` of `str`

    :param lineage: Certificate lineage object. Defaults to `None`
    :type lineage: storage.RenewableCert

    :returns: `None`
    :rtype: None

    """
    path_provider: Union[storage.RenewableCert,
                         configuration.NamespaceConfig] = lineage if lineage else config
    assert path_provider.cert_path is not None

    domains, ip_addresses = san.split(sans)
    if len(ip_addresses) > 0:
        # Our apache and nginx plugins are currently relying on this check for a user friendly error
        # message about their lack of support for IP certificates. If you're removing this check,
        # please check that the plugins can process IP addresses.
        raise errors.ConfigurationError("Enhancements not supported for IP address certificates")

    le_client.deploy_certificate(domains, path_provider.key_path, path_provider.cert_path,
                                 path_provider.chain_path, path_provider.fullchain_path)

    le_client.enhance_config(domains, path_provider.chain_path)


def install(config: configuration.NamespaceConfig,
            plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Install a previously obtained cert in a server.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :returns: `None` or the error message
    :rtype: None or str

    """
    # XXX: Update for renewer/RenewableCert
    # FIXME: be consistent about whether errors are raised or returned from
    # this function ...

    try:
        installer, _ = plug_sel.choose_configurator_plugins(config, plugins, "install")
    except errors.PluginSelectionError as e:
        return str(e)

    custom_cert = (config.key_path and config.cert_path)
    if not config.certname and not custom_cert:
        certname_question = "Which certificate would you like to install?"
        config.certname = cert_manager.get_certnames(
            config, "install", allow_multiple=False,
            custom_prompt=certname_question)[0]

    if not enhancements.are_supported(config, installer):
        raise errors.NotSupportedError("One or more of the requested enhancements "
                                       "are not supported by the selected installer")
    # If cert-path is defined, populate missing (ie. not overridden) values.
    # Unfortunately this can't be done in argument parser, as certificate
    # manager needs the access to renewal directory paths
    if config.certname:
        config = _populate_from_certname(config)
    elif enhancements.are_requested(config):
        # Preflight config check
        raise errors.ConfigurationError("One or more of the requested enhancements "
                                        "require --cert-name to be provided")

    if config.key_path and config.cert_path:
        _check_certificate_and_key(config)
        sans, _ = _find_sans_or_certname(config, installer)
        le_client = _init_le_client(config, authenticator=None, installer=installer)
        _install_cert(config, le_client, sans)
    else:
        raise errors.ConfigurationError("Path to certificate or key was not defined. "
            "If your certificate is managed by Certbot, please use --cert-name "
            "to define which certificate you would like to install.")

    if enhancements.are_requested(config):
        # In the case where we don't have certname, we have errored out already
        lineage = cert_manager.lineage_for_certname(config, config.certname)
        domains, ip_addresses = san.split(sans)
        if ip_addresses:
            raise TypeError("enhancements not supported for IP address certificates")
        domains_str = [d.dns_name for d in domains]
        enhancements.enable(lineage, domains_str, installer, config)

    return None


def _populate_from_certname(config: configuration.NamespaceConfig) -> configuration.NamespaceConfig:
    """Helper function for install to populate missing config values from lineage
    defined by --cert-name."""

    lineage = cert_manager.lineage_for_certname(config, config.certname)
    if not lineage:
        return config
    if not config.key_path:
        config.key_path = lineage.key_path
    if not config.cert_path:
        config.cert_path = lineage.cert_path
    if not config.chain_path:
        config.chain_path = lineage.chain_path
    if not config.fullchain_path:
        config.fullchain_path = lineage.fullchain_path
    return config


def _check_certificate_and_key(config: configuration.NamespaceConfig) -> None:
    if not os.path.isfile(filesystem.realpath(config.cert_path)):
        raise errors.ConfigurationError("Error while reading certificate from path "
                                        "{0}".format(config.cert_path))
    if not os.path.isfile(filesystem.realpath(config.key_path)):
        raise errors.ConfigurationError("Error while reading private key from path "
                                        "{0}".format(config.key_path))


def plugins_cmd(config: configuration.NamespaceConfig,
                plugins: plugins_disco.PluginsRegistry) -> None:
    """List server software plugins.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    """
    logger.debug("Expected interfaces: %s", config.ifaces)

    ifaces = [] if config.ifaces is None else config.ifaces
    filtered = plugins.visible().ifaces(ifaces)
    logger.debug("Filtered plugins: %r", filtered)

    notify = functools.partial(display_util.notification, pause=False)
    if not config.init and not config.prepare:
        notify(str(filtered))
        return

    filtered.init(config)
    logger.debug("Filtered plugins: %r", filtered)

    if not config.prepare:
        notify(str(filtered))
        return

    filtered.prepare()
    available = filtered.available()
    logger.debug("Prepared plugins: %s", available)
    notify(str(available))


def enhance(config: configuration.NamespaceConfig,
            plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Add security enhancements to existing configuration

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :returns: `None` or a string indicating an error
    :rtype: None or str

    """
    supported_enhancements = ["hsts", "redirect", "uir", "staple"]
    # Check that at least one enhancement was requested on command line
    oldstyle_enh = any(getattr(config, enh) for enh in supported_enhancements)
    if not enhancements.are_requested(config) and not oldstyle_enh:
        msg = ("Please specify one or more enhancement types to configure. To list "
               "the available enhancement types, run:\n\n%s --help enhance\n")
        logger.error(msg, cli.cli_command)
        raise errors.MisconfigurationError("No enhancements requested, exiting.")

    try:
        installer, _ = plug_sel.choose_configurator_plugins(config, plugins, "enhance")
    except errors.PluginSelectionError as e:
        return str(e)

    if not enhancements.are_supported(config, installer):
        raise errors.NotSupportedError("One or more of the requested enhancements "
                                       "are not supported by the selected installer")

    certname_question = ("Which certificate would you like to use to enhance "
                         "your configuration?")
    config.certname = cert_manager.get_certnames(
        config, "enhance", allow_multiple=False,
        custom_prompt=certname_question)[0]
    cert_sans = cert_manager.sans_for_certname(config, config.certname)
    if cert_sans is None:
        raise errors.Error("Could not find the list of domains for the given certificate name.")
    cert_domains, ip_addresses = san.split(cert_sans)
    if len(ip_addresses) > 0:
        # Our apache and nginx plugins are currently relying on this check for a user friendly error
        # message about their lack of support for IP certificates. If you're removing this check,
        # please check that the plugins can process IP addresses.
        raise errors.ConfigurationError("Enhancements not supported for IP address certificates")

    if config.noninteractive_mode:
        domains = cert_domains
    else:
        domain_question = ("Which domain names would you like to enable the "
                           "selected enhancements for?")
        domain_strs = display_ops.choose_values(list(map(str, cert_domains)), domain_question)
        if not domain_strs:
            raise errors.Error("User cancelled the domain selection. No domains "
                               "defined, exiting.")
        domains = list(map(san.DNSName, domain_strs))

    lineage = cert_manager.lineage_for_certname(config, config.certname)
    if not lineage:
        raise errors.Error("Could not find the lineage for the given certificate name.")
    if not config.chain_path:
        config.chain_path = lineage.chain_path
    if oldstyle_enh:
        le_client = _init_le_client(config, authenticator=None, installer=installer)
        le_client.enhance_config(domains, config.chain_path, redirect_default=False)
    if enhancements.are_requested(config):
        enhancements.enable(lineage, [d.dns_name for d in domains], installer, config)

    return None


def rollback(config: configuration.NamespaceConfig, plugins: plugins_disco.PluginsRegistry) -> None:
    """Rollback server configuration changes made during install.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    """
    client.rollback(config.installer, config.checkpoints, config, plugins)


def delete(config: configuration.NamespaceConfig,
           unused_plugins: plugins_disco.PluginsRegistry) -> None:
    """Delete a certificate

    Use the information in the config file to delete an existing
    lineage.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    """
    cert_manager.delete(config)


def certificates(config: configuration.NamespaceConfig,
                 unused_plugins: plugins_disco.PluginsRegistry) -> None:
    """Display information about certs configured with Certbot

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    """
    cert_manager.certificates(config)


def revoke(config: configuration.NamespaceConfig,
           unused_plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Revoke a previously obtained certificate.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None` or string indicating error in case of error
    :rtype: None or str

    """
    # For user-agent construction
    config.installer = config.authenticator = None

    if config.cert_path is None and config.certname:
        # When revoking via --cert-name, take the cert path and server from renewalparams
        lineage = storage.RenewableCert(
            storage.renewal_file_for_certname(config, config.certname), config)
        config.cert_path = lineage.cert_path
        # --server takes priority over lineage.server
        if lineage.server and not config.set_by_user("server"):
            config.server = lineage.server
    elif not config.cert_path or (config.cert_path and config.certname):
        # intentionally not supporting --cert-path & --cert-name together,
        # to avoid dealing with mismatched values
        raise errors.Error("Error! Exactly one of --cert-path or --cert-name must be specified!")

    if config.key_path is not None:  # revocation by cert key
        logger.debug("Revoking %s using certificate key %s",
                     config.cert_path, config.key_path)
        crypto_util.verify_cert_matches_priv_key(config.cert_path, config.key_path)
        with open(config.key_path, 'rb') as f:
            key = jose.JWK.load(f.read())
        acme = client.create_acme_client(config, key)
    else:  # revocation by account key
        logger.debug("Revoking %s using Account Key", config.cert_path)
        acc, _ = _determine_account(config)
        acme = client.create_acme_client(config, acc.key, acc.regr)

    with open(config.cert_path, 'rb') as f:
        cert = x509.load_pem_x509_certificate(f.read())
    logger.debug("Reason code for revocation: %s", config.reason)
    try:
        acme.revoke(cert, config.reason)
        _delete_if_appropriate(config)
    except acme_errors.ClientError as e:
        return str(e)

    display_ops.success_revocation(config.cert_path)
    return None


def run(config: configuration.NamespaceConfig,
        plugins: plugins_disco.PluginsRegistry) -> Optional[str]:
    """Obtain a certificate and install.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    """
    # TODO: Make run as close to auth + install as possible
    # Possible difficulties: config.csr was hacked into auth
    try:
        installer, authenticator = plug_sel.choose_configurator_plugins(config, plugins, "run")
    except errors.PluginSelectionError as e:
        return str(e)

    if config.must_staple and installer and "staple-ocsp" not in installer.supported_enhancements():
        raise errors.NotSupportedError(
            "Must-Staple extension requested, but OCSP stapling is not supported by the selected "
            f"installer ({config.installer})\n\n"
            "You can either:\n"
            " * remove the --must-staple option from the command line and obtain a certificate "
            "without the Must-Staple extension, or;\n"
            " * use the `certonly` subcommand and manually install the certificate into the  "
            "intended service (e.g. webserver). You must also then manually enable OCSP stapling, "
            "as it is required for certificates with the Must-Staple extension to "
            "function properly.\n"
            " * choose a different installer plugin (such as --nginx or --apache), if possible."
        )

    # Preflight check for enhancement support by the selected installer
    if not enhancements.are_supported(config, installer):
        raise errors.NotSupportedError("One or more of the requested enhancements "
                                       "are not supported by the selected installer")

    # TODO: Handle errors from _init_le_client?
    le_client = _init_le_client(config, authenticator, installer)

    sans, certname = _find_sans_or_certname(config, installer)

    domains, ip_addresses = san.split(sans)
    if ip_addresses:
        raise errors.Error("installation of IP address certificate not supported")

    should_get_cert, lineage = _find_cert(config, sans, certname)

    new_lineage = lineage
    if should_get_cert:
        new_lineage = _get_and_save_cert(le_client, config, sans,
            certname, lineage)

    cert_path = new_lineage.cert_path if new_lineage else None
    fullchain_path = new_lineage.fullchain_path if new_lineage else None
    key_path = new_lineage.key_path if new_lineage else None

    if should_get_cert:
        _report_new_cert(config, cert_path, fullchain_path, key_path)

    # The installer error, if any, is being stored as a value here, in order to first print
    # relevant advice in a nice way, before re-raising the error for normal processing.
    installer_err: Optional[errors.Error] = None
    try:
        _install_cert(config, le_client, sans, new_lineage)

        if enhancements.are_requested(config) and new_lineage:
            enhancements.enable(new_lineage, [d.dns_name for d in domains], installer, config)

        sans_strs = list(map(str, sans))
        if lineage is None or not should_get_cert:
            display_ops.success_installation(sans_strs)
        else:
            display_ops.success_renewal(sans_strs)
    except errors.Error as e:
        installer_err = e
    finally:
        _report_next_steps(config, installer_err, new_lineage,
                           new_or_renewed_cert=should_get_cert)
        # If the installer did fail, re-raise the error to bail out
        if installer_err:
            raise installer_err

    _suggest_donation_if_appropriate(config)
    eff.handle_subscription(config, le_client.account)
    return None


def _csr_get_and_save_cert(config: configuration.NamespaceConfig,
                           le_client: client.Client) -> tuple[
                           Optional[str], Optional[str], Optional[str]]:
    """Obtain a cert using a user-supplied CSR

    This works differently in the CSR case (for now) because we don't
    have the privkey, and therefore can't construct the files for a lineage.
    So we just save the cert & chain to disk :/

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param client: Client object
    :type client: client.Client

    :returns: `cert_path`, `chain_path` and `fullchain_path` as absolute
              paths to the actual files, or None for each if it's a dry-run.
    :rtype: `tuple` of `str`

    """
    util_csr = config.actual_csr
    x509_req = x509.load_pem_x509_csr(util_csr.data)
    domains, ip_addresses = san.from_x509(x509_req.subject, x509_req.extensions)
    display_util.notify(
        "{action} for {sans}".format(
            action="Simulating a certificate request" if config.dry_run else
                    "Requesting a certificate",
            sans=internal_display_util.summarize_sans(san.join(domains, ip_addresses))
        )
    )
    cert, chain = le_client.obtain_certificate_from_csr(util_csr)
    if config.dry_run:
        logger.debug(
            "Dry run: skipping saving certificate to %s", config.cert_path)
        return None, None, None
    cert_path, chain_path, fullchain_path = le_client.save_certificate(
        cert, chain, os.path.normpath(config.cert_path),
        os.path.normpath(config.chain_path), os.path.normpath(config.fullchain_path))
    return cert_path, chain_path, fullchain_path


def renew_cert(config: configuration.NamespaceConfig, plugins: plugins_disco.PluginsRegistry,
               lineage: storage.RenewableCert) -> None:
    """Renew & save an existing cert. Do not install it.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :param lineage: Certificate lineage object
    :type lineage: storage.RenewableCert

    :returns: `None`
    :rtype: None

    :raises errors.PluginSelectionError: MissingCommandlineFlag if supplied parameters do not pass

    """
    # installers are used in auth mode to determine domain names
    installer, auth = plug_sel.choose_configurator_plugins(config, plugins, "certonly")
    le_client = _init_le_client(config, auth, installer)

    renewed_lineage = _get_and_save_cert(le_client, config, lineage=lineage)

    if not renewed_lineage:
        raise errors.Error("An existing certificate for the given name could not be found.")

    if installer and not config.dry_run:
        # In case of a renewal, reload server to pick up new certificate.
        updater.run_renewal_deployer(config, renewed_lineage, installer)
        display_util.notify(f"Reloading {config.installer} server after certificate renewal")
        installer.restart()


def certonly(config: configuration.NamespaceConfig, plugins: plugins_disco.PluginsRegistry) -> None:
    """Authenticate & obtain cert, but do not install it.

    This implements the 'certonly' subcommand.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    :raises errors.Error: If specified plugin could not be used

    """
    # SETUP: Select plugins and construct a client instance
    # installers are used in auth mode to determine domain names
    installer, auth = plug_sel.choose_configurator_plugins(config, plugins, "certonly")
    le_client = _init_le_client(config, auth, installer)

    if config.csr:
        cert_path, chain_path, fullchain_path = _csr_get_and_save_cert(config, le_client)
        _csr_report_new_cert(config, cert_path, chain_path, fullchain_path)
        _report_next_steps(config, None, None, new_or_renewed_cert=not config.dry_run)
        _suggest_donation_if_appropriate(config)
        eff.handle_subscription(config, le_client.account)
        return

    sans, certname = _find_sans_or_certname(config, installer)
    should_get_cert, lineage = _find_cert(config, sans, certname)

    if not should_get_cert:
        display_util.notification("Certificate not yet due for renewal; no action taken.",
                                     pause=False)
        return

    lineage = _get_and_save_cert(le_client, config, sans, certname, lineage)

    # If a new cert was issued and we were passed an installer, we can safely
    # run `installer.restart()` to load the newly issued certificate
    installer_err: Optional[errors.Error] = None
    if lineage and installer and not config.dry_run:
        logger.info("Reloading %s server after certificate issuance", config.installer)
        try:
            installer.restart()
        except errors.Error as e:
            installer_err = e

    cert_path = lineage.cert_path if lineage else None
    fullchain_path = lineage.fullchain_path if lineage else None
    key_path = lineage.key_path if lineage else None
    _report_new_cert(config, cert_path, fullchain_path, key_path)
    _report_next_steps(config, installer_err, lineage,
                       new_or_renewed_cert=should_get_cert and not config.dry_run)
    if installer_err:
        raise installer_err
    _suggest_donation_if_appropriate(config)
    eff.handle_subscription(config, le_client.account)


def renew(config: configuration.NamespaceConfig,
          unused_plugins: plugins_disco.PluginsRegistry) -> None:
    """Renew previously-obtained certificates.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param unused_plugins: List of plugins (deprecated)
    :type unused_plugins: plugins_disco.PluginsRegistry

    :returns: `None`
    :rtype: None

    """
    renewal.handle_renewal_request(config)


def make_or_verify_needed_dirs(config: configuration.NamespaceConfig) -> None:
    """Create or verify existence of config, work, and hook directories.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :returns: `None`
    :rtype: None

    """
    util.set_up_core_dir(config.config_dir, constants.CONFIG_DIRS_MODE, config.strict_permissions)

    # Ensure the working directory has the expected mode, even under stricter umask settings
    with filesystem.temp_umask(0o022):
        util.set_up_core_dir(config.work_dir, constants.CONFIG_DIRS_MODE, config.strict_permissions)

    hook_dirs = (config.renewal_pre_hooks_dir,
                 config.renewal_deploy_hooks_dir,
                 config.renewal_post_hooks_dir,)
    for hook_dir in hook_dirs:
        util.make_or_verify_dir(hook_dir, strict=config.strict_permissions)


def _report_reconfigure_results(renewal_file: str, orig_renewal_conf: configobj.ConfigObj) -> None:
    """Reports the outcome of certificate renewal reconfiguration to the user.

    :param renewal_file: Path to the cert's renewal file
    :type renewal_file: str

    :param orig_renewal_conf: Loaded original renewal configuration
    :type orig_renewal_conf: configobj.ConfigObj

    :returns: `None`
    :rtype: None

    """
    try:
        final_renewal_conf = configobj.ConfigObj(
            renewal_file, encoding='utf-8', default_encoding='utf-8')
    except configobj.ConfigObjError:
        raise errors.CertStorageError(
            f'error parsing {renewal_file}')

    orig_renewal_params = orig_renewal_conf['renewalparams']
    final_renewal_params = final_renewal_conf['renewalparams']

    if final_renewal_params == orig_renewal_params:
        success_message = '\nNo changes were made to the renewal configuration.'
    else:
        success_message = '\nSuccessfully updated configuration.' + \
                          '\nChanges will apply when the certificate renews.'

    display_util.notify(success_message)


def reconfigure(config: configuration.NamespaceConfig,
          plugins: plugins_disco.PluginsRegistry) -> None:
    """Allow the user to set new configuration options for an existing certificate without
       forcing renewal. This can be used for things like authenticator, installer, and hooks,
       but not for the domains on the cert, since those are only saved in the cert.

    :param config: Configuration object
    :type config: configuration.NamespaceConfig

    :param plugins: List of plugins
    :type plugins: plugins_disco.PluginsRegistry

    :raises errors.Error: if the dry run fails
    :raises errors.ConfigurationError: if certificate could not be loaded

    """

    if config.domains:
        raise errors.ConfigurationError("You have specified domains, but this function cannot "
            "be used to modify the domains in a certificate. If you would like to do so, follow "
            "the instructions at https://certbot.org/change-cert-domain. Otherwise, remove the "
            "domains from the command to continue reconfiguring. You can specify which certificate "
            "you want on the command line with flag --cert-name instead.")
    # While we could technically allow domains to be used to specify the certificate in addition to
    # --cert-name, there's enough complexity with matching certs to domains that it's not worth it,
    # to say nothing of the difficulty in explaining what exactly this subcommand can modify


    # To make sure that the requested changes work, we're going to do a dry run, and only save
    # upon success. First, modify the config as the user requested.
    if not config.certname:
        certname_question = "Which certificate would you like to reconfigure?"
        config.certname = cert_manager.get_certnames(
            config, "reconfigure", allow_multiple=False,
            custom_prompt=certname_question)[0]

    certname = config.certname

    try:
        renewal_file = storage.renewal_file_for_certname(config, certname)
    except errors.CertStorageError:
        raise errors.ConfigurationError(f"An existing certificate with name {certname} could not "
            "be found. Run `certbot certificates` to list available certificates.")

    # figure this out before we modify config
    if config.deploy_hook and not config.run_deploy_hooks:
        msg = ("You are attempting to set a --deploy-hook. Would you like Certbot to run deploy "
               "hooks when it performs a dry run with the new settings? This will run all "
               "relevant deploy hooks, including directory hooks, unless --no-directory-hooks "
               "is set. This will use the current active certificate, and not the temporary test "
               "certificate acquired during the dry run.")
        config.run_deploy_hooks = display_util.yesno(msg,"Run deploy hooks",
            "Do not run deploy hooks", default=False)

    # cache previous version for later comparison
    try:
        orig_renewal_conf = configobj.ConfigObj(
            renewal_file, encoding='utf-8', default_encoding='utf-8')
    except configobj.ConfigObjError:
        raise errors.CertStorageError(
            f"error parsing {renewal_file}")

    lineage_config = copy.deepcopy(config)
    try:
        renewal_candidate = renewal.reconstitute(lineage_config, renewal_file)
    except Exception as e:  # pylint: disable=broad-except
        raise errors.ConfigurationError(f"Renewal configuration file {renewal_file} "
            f"(cert: {certname}) produced an unexpected error: {e}.")
    if not renewal_candidate:
        raise errors.ConfigurationError("Could not load certificate. See logs for errors.")

    renewalparams = orig_renewal_conf['renewalparams']
    # If server was set but hasn't changed and no account is loaded,
    # load the old account because reconstitute won't have
    if lineage_config.set_by_user('server') and lineage_config.server == renewalparams['server']\
        and lineage_config.account is None:
        lineage_config.account = renewalparams['account']
    for param in ('account', 'server',):
        if getattr(lineage_config, param) != renewalparams.get(param):
            msg = ("Using reconfigure to change the ACME account or server is not supported. "
                   "If you would like to do so, use renew with the --force-renewal flag instead "
                   "of reconfigure. Note that doing so will count against any rate limits. For "
                   "more information on this method, see "
                   "https://certbot.org/renew-reconfiguration")
            raise errors.ConfigurationError(msg)

    # this is where lineage_config gets fully filled out (e.g. --apache will set auth and installer)
    installer, auth = plug_sel.choose_configurator_plugins(lineage_config, plugins, "certonly")

    # make a deep copy of lineage_config because we're about to modify it for a test dry run
    dry_run_lineage_config = copy.deepcopy(lineage_config)

    # we also set noninteractive_mode to more accurately simulate renewal (since `certbot renew`
    # implies noninteractive mode) and to avoid prompting the user as changes made to
    # dry_run_lineage_config beyond this point will not be applied to the original lineage_config
    dry_run_lineage_config.noninteractive_mode = True
    dry_run_lineage_config.dry_run = True
    cli.set_test_server_options("reconfigure", dry_run_lineage_config)

    le_client = _init_le_client(dry_run_lineage_config, auth, installer)

    # renews cert as dry run to test that the new values are ok
    # at this point, renewal_candidate.configuration has the old values, but will use
    # the values from lineage_config when doing the dry run
    _get_and_save_cert(le_client, dry_run_lineage_config, certname=certname,
        lineage=renewal_candidate)

    # this function will update lineage.configuration with the new values, and save it to disk
    # use the pre-dry-run version
    renewal_candidate.save_new_config_values(lineage_config)

    _report_reconfigure_results(renewal_file, orig_renewal_conf)


@contextmanager
def make_displayer(config: configuration.NamespaceConfig
                   ) -> Generator[Union[display_obj.NoninteractiveDisplay,
                                        display_obj.FileDisplay], None, None]:
    """Creates a display object appropriate to the flags in the supplied config.

    :param config: Configuration object

    :returns: Display object

    """
    displayer: Union[None, display_obj.NoninteractiveDisplay,
                     display_obj.FileDisplay] = None
    devnull: Optional[IO] = None

    if config.quiet:
        config.noninteractive_mode = True
        devnull = open(os.devnull, "w")  # pylint: disable=consider-using-with
        displayer = display_obj.NoninteractiveDisplay(devnull)
    elif config.noninteractive_mode:
        displayer = display_obj.NoninteractiveDisplay(sys.stdout)
    else:
        displayer = display_obj.FileDisplay(
            sys.stdout, config.force_interactive)

    try:
        yield displayer
    finally:
        if devnull:
            devnull.close()


def main(cli_args: Optional[list[str]] = None) -> Optional[Union[str, int]]:
    """Run Certbot.

    :param cli_args: command line to Certbot, defaults to ``sys.argv[1:]``
    :type cli_args: `list` of `str`

    :returns: value for `sys.exit` about the exit status of Certbot
    :rtype: `str` or `int` or `None`

    """
    if not cli_args:
        cli_args = sys.argv[1:]

    log.pre_arg_parse_setup()

    if os.environ.get('CERTBOT_SNAPPED') == 'True':
        cli_args = snap_config.prepare_env(cli_args)

    plugins = plugins_disco.PluginsRegistry.find_all()
    logger.debug("certbot version: %s", certbot.__version__)
    logger.debug("Location of certbot entry point: %s", sys.argv[0])
    # do not log `config`, as it contains sensitive data (e.g. revoke --key)!
    logger.debug("Arguments: %r", cli_args)
    logger.debug("Discovered plugins: %r", plugins)

    # Some releases of Windows require escape sequences to be enable explicitly
    misc.prepare_virtual_console()

    # note: arg parser internally handles --help (and exits afterwards)
    config = cli.prepare_and_parse_args(plugins, cli_args)

    # On windows, shell without administrative right cannot create symlinks required by certbot.
    # So we check the rights before continuing.
    misc.raise_for_non_administrative_windows_rights()

    try:
        log.post_arg_parse_setup(config)
        make_or_verify_needed_dirs(config)
    except errors.Error:
        # Let plugins_cmd be run as un-privileged user.
        if config.func != plugins_cmd:  # pylint: disable=comparison-with-callable
            raise

    with make_displayer(config) as displayer:
        display_obj.set_display(displayer)

        return config.func(config, plugins)



"""Certbot display.

This module (`certbot.display.util`) or its companion `certbot.display.ops`
should be used whenever:

- Displaying status information to the user on the terminal
- Collecting information from the user via prompts

Other messages can use the `logging` module. See `log.py`.

"""
from typing import Optional
from typing import Union

from certbot._internal.display import obj

# These constants are defined this way to make them easier to document with
# Sphinx and to not couple our public docstrings to our internal ones.
OK = obj.OK
"""Display exit code indicating user acceptance."""

CANCEL = obj.CANCEL
"""Display exit code for a user canceling the display."""

WIDTH = 72

def notify(msg: str) -> None:
    """Display a basic status message.

    :param str msg: message to display

    """
    obj.get_display().notification(msg, pause=False, decorate=False, wrap=False)


def notification(message: str, pause: bool = True, wrap: bool = True,
                 force_interactive: bool = False, decorate: bool = True) -> None:
    """Displays a notification and waits for user acceptance.

    :param str message: Message to display
    :param bool pause: Whether or not the program should pause for the
        user's confirmation
    :param bool wrap: Whether or not the application should wrap text
    :param bool force_interactive: True if it's safe to prompt the user
        because it won't cause any workflow regressions
    :param bool decorate: Whether to surround the message with a
        decorated frame

    """
    obj.get_display().notification(message, pause=pause, wrap=wrap,
                                   force_interactive=force_interactive, decorate=decorate)


def menu(message: str, choices: Union[list[str], list[tuple[str, str]]],
         default: Optional[int] = None, cli_flag: Optional[str] = None,
         force_interactive: bool = False) -> tuple[str, int]:
    """Display a menu.

    .. todo:: This doesn't enable the help label/button (I wasn't sold on
        any interface I came up with for this). It would be a nice feature.

    :param str message: title of menu
    :param choices: Menu lines, len must be > 0
    :type choices: list of tuples (tag, item) or
        list of descriptions (tags will be enumerated)
    :param default: default value to return, if interaction is not possible
    :param str cli_flag: option used to set this value with the CLI
    :param bool force_interactive: True if it's safe to prompt the user
        because it won't cause any workflow regressions

    :returns: tuple of (`code`, `index`) where
        `code` - str display exit code
        `index` - int index of the user's selection

    :rtype: tuple

    """
    return obj.get_display().menu(message, choices, default=default, cli_flag=cli_flag,
                                  force_interactive=force_interactive)


def input_text(message: str, default: Optional[str] = None, cli_flag: Optional[str] = None,
               force_interactive: bool = False) -> tuple[str, str]:
    """Accept input from the user.

    :param str message: message to display to the user
    :param default: default value to return, if interaction is not possible
    :param str cli_flag: option used to set this value with the CLI
    :param bool force_interactive: True if it's safe to prompt the user
        because it won't cause any workflow regressions

    :returns: tuple of (`code`, `input`) where
        `code` - str display exit code
        `input` - str of the user's input
    :rtype: tuple

    """
    return obj.get_display().input(message, default=default, cli_flag=cli_flag,
                                   force_interactive=force_interactive)


def yesno(message: str, yes_label: str = "Yes", no_label: str = "No",
          default: Optional[bool] = None, cli_flag: Optional[str] = None,
          force_interactive: bool = False) -> bool:
    """Query the user with a yes/no question.

    Yes and No label must begin with different letters, and must contain at
    least one letter each.

    :param str message: question for the user
    :param str yes_label: Label of the "Yes" parameter
    :param str no_label: Label of the "No" parameter
    :param default: default value to return, if interaction is not possible
    :param str cli_flag: option used to set this value with the CLI
    :param bool force_interactive: True if it's safe to prompt the user
        because it won't cause any workflow regressions

    :returns: True for "Yes", False for "No"
    :rtype: bool

    """
    return obj.get_display().yesno(message, yes_label=yes_label, no_label=no_label, default=default,
                                   cli_flag=cli_flag, force_interactive=force_interactive)


def checklist(message: str, tags: list[str], default: Optional[list[str]] = None,
              cli_flag: Optional[str] = None,
              force_interactive: bool = False) -> tuple[str, list[str]]:
    """Display a checklist.

    :param str message: Message to display to user
    :param list tags: `str` tags to select, len(tags) > 0
    :param default: default value to return, if interaction is not possible
    :param str cli_flag: option used to set this value with the CLI
    :param bool force_interactive: True if it's safe to prompt the user
        because it won't cause any workflow regressions

    :returns: tuple of (`code`, `tags`) where
        `code` - str display exit code
        `tags` - list of selected tags
    :rtype: tuple

    """
    return obj.get_display().checklist(message, tags, default=default, cli_flag=cli_flag,
                                       force_interactive=force_interactive)


def directory_select(message: str, default: Optional[str] = None, cli_flag: Optional[str] = None,
                     force_interactive: bool = False) -> tuple[str, str]:
    """Display a directory selection screen.

    :param str message: prompt to give the user
    :param default: default value to return, if interaction is not possible
    :param str cli_flag: option used to set this value with the CLI
    :param bool force_interactive: True if it's safe to prompt the user
        because it won't cause any workflow regressions

    :returns: tuple of the form (`code`, `string`) where
        `code` - display exit code
        `string` - input entered by the user

    """
    return obj.get_display().directory_select(message, default=default, cli_flag=cli_flag,
                                              force_interactive=force_interactive)


def assert_valid_call(prompt: str, default: str, cli_flag: str, force_interactive: bool) -> None:
    """Verify that provided arguments is a valid display call.

    :param str prompt: prompt for the user
    :param default: default answer to prompt
    :param str cli_flag: command line option for setting an answer
        to this question
    :param bool force_interactive: if interactivity is forced

    """
    msg = "Invalid display call for this prompt:\n{0}".format(prompt)
    if cli_flag:
        msg += ("\nYou can set an answer to "
                "this prompt with the {0} flag".format(cli_flag))
    assert default is not None or force_interactive, msg



"""Tests for DualParserNode implementation"""
import unittest
from unittest import mock

from certbot._internal.plugins.apache import assertions
from certbot._internal.plugins.apache import augeasparser
from certbot._internal.plugins.apache import dualparser
import pytest


class DualParserNodeTest(unittest.TestCase):  # pylint: disable=too-many-public-methods
    """DualParserNode tests"""

    def setUp(self):  # pylint: disable=arguments-differ
        parser_mock = mock.MagicMock()
        parser_mock.aug.match.return_value = []
        parser_mock.get_arg.return_value = []
        self.metadata = {"augeasparser": parser_mock, "augeaspath": "/invalid", "ac_ast": None}
        self.block = dualparser.DualBlockNode(name="block",
                                              ancestor=None,
                                              filepath="/tmp/something",
                                              metadata=self.metadata)
        self.block_two = dualparser.DualBlockNode(name="block",
                                                  ancestor=self.block,
                                                  filepath="/tmp/something",
                                                  metadata=self.metadata)
        self.directive = dualparser.DualDirectiveNode(name="directive",
                                                      ancestor=self.block,
                                                      filepath="/tmp/something",
                                                      metadata=self.metadata)
        self.comment = dualparser.DualCommentNode(comment="comment",
                                                  ancestor=self.block,
                                                  filepath="/tmp/something",
                                                  metadata=self.metadata)

    def test_create_with_precreated(self):
        cnode = dualparser.DualCommentNode(comment="comment",
                                           ancestor=self.block,
                                           filepath="/tmp/something",
                                           primary=self.comment.secondary,
                                           secondary=self.comment.primary)
        dnode = dualparser.DualDirectiveNode(name="directive",
                                             ancestor=self.block,
                                             filepath="/tmp/something",
                                             primary=self.directive.secondary,
                                             secondary=self.directive.primary)
        bnode = dualparser.DualBlockNode(name="block",
                                         ancestor=self.block,
                                         filepath="/tmp/something",
                                         primary=self.block.secondary,
                                         secondary=self.block.primary)
        # Switched around
        assert cnode.primary == self.comment.secondary
        assert cnode.secondary == self.comment.primary
        assert dnode.primary == self.directive.secondary
        assert dnode.secondary == self.directive.primary
        assert bnode.primary == self.block.secondary
        assert bnode.secondary == self.block.primary

    def test_set_params(self):
        params = ("first", "second")
        self.directive.primary.set_parameters = mock.Mock()
        self.directive.secondary.set_parameters = mock.Mock()
        self.directive.set_parameters(params)
        assert self.directive.primary.set_parameters.called is True
        assert self.directive.secondary.set_parameters.called is True

    def test_set_parameters(self):
        pparams = mock.MagicMock()
        sparams = mock.MagicMock()
        pparams.parameters = ("a", "b")
        sparams.parameters = ("a", "b")
        self.directive.primary.set_parameters = pparams
        self.directive.secondary.set_parameters = sparams
        self.directive.set_parameters(("param", "seq"))
        assert pparams.called is True
        assert sparams.called is True

    def test_delete_child(self):
        pdel = mock.MagicMock()
        sdel = mock.MagicMock()
        self.block.primary.delete_child = pdel
        self.block.secondary.delete_child = sdel
        self.block.delete_child(self.comment)
        assert pdel.called is True
        assert sdel.called is True

    def test_unsaved_files(self):
        puns = mock.MagicMock()
        suns = mock.MagicMock()
        puns.return_value = assertions.PASS
        suns.return_value = assertions.PASS
        self.block.primary.unsaved_files = puns
        self.block.secondary.unsaved_files = suns
        self.block.unsaved_files()
        assert puns.called is True
        assert suns.called is True

    def test_getattr_equality(self):
        self.directive.primary.variableexception = "value"
        self.directive.secondary.variableexception = "not_value"
        with pytest.raises(AssertionError):
            _ = self.directive.variableexception

        self.directive.primary.variable = "value"
        self.directive.secondary.variable = "value"
        try:
            self.directive.variable
        except AssertionError: # pragma: no cover
            self.fail("getattr check raised an AssertionError where it shouldn't have")

    def test_parsernode_dirty_assert(self):
        # disable assertion pass
        self.comment.primary.comment = "value"
        self.comment.secondary.comment = "value"
        self.comment.primary.filepath = "x"
        self.comment.secondary.filepath = "x"

        self.comment.primary.dirty = False
        self.comment.secondary.dirty = True
        with pytest.raises(AssertionError):
            assertions.assertEqual(self.comment.primary, self.comment.secondary)

    def test_parsernode_filepath_assert(self):
        # disable assertion pass
        self.comment.primary.comment = "value"
        self.comment.secondary.comment = "value"

        self.comment.primary.filepath = "first"
        self.comment.secondary.filepath = "second"
        with pytest.raises(AssertionError):
            assertions.assertEqual(self.comment.primary, self.comment.secondary)

    def test_add_child_block(self):
        mock_first = mock.MagicMock(return_value=self.block.primary)
        mock_second = mock.MagicMock(return_value=self.block.secondary)
        self.block.primary.add_child_block = mock_first
        self.block.secondary.add_child_block = mock_second
        self.block.add_child_block("Block")
        assert mock_first.called is True
        assert mock_second.called is True

    def test_add_child_directive(self):
        mock_first = mock.MagicMock(return_value=self.directive.primary)
        mock_second = mock.MagicMock(return_value=self.directive.secondary)
        self.block.primary.add_child_directive = mock_first
        self.block.secondary.add_child_directive = mock_second
        self.block.add_child_directive("Directive")
        assert mock_first.called is True
        assert mock_second.called is True

    def test_add_child_comment(self):
        mock_first = mock.MagicMock(return_value=self.comment.primary)
        mock_second = mock.MagicMock(return_value=self.comment.secondary)
        self.block.primary.add_child_comment = mock_first
        self.block.secondary.add_child_comment = mock_second
        self.block.add_child_comment("Comment")
        assert mock_first.called is True
        assert mock_second.called is True

    def test_find_comments(self):
        pri_comments = [augeasparser.AugeasCommentNode(comment="some comment",
                                                       ancestor=self.block,
                                                       filepath="/path/to/whatever",
                                                       metadata=self.metadata)]
        sec_comments = [augeasparser.AugeasCommentNode(comment=assertions.PASS,
                                                       ancestor=self.block,
                                                       filepath=assertions.PASS,
                                                       metadata=self.metadata)]
        find_coms_primary = mock.MagicMock(return_value=pri_comments)
        find_coms_secondary = mock.MagicMock(return_value=sec_comments)
        self.block.primary.find_comments = find_coms_primary
        self.block.secondary.find_comments = find_coms_secondary

        dcoms = self.block.find_comments("comment")
        p_dcoms = [d.primary for d in dcoms]
        s_dcoms = [d.secondary for d in dcoms]
        p_coms = self.block.primary.find_comments("comment")
        s_coms = self.block.secondary.find_comments("comment")
        # Check that every comment response is represented in the list of
        # DualParserNode instances.
        for p in p_dcoms:
            assert p in p_coms
        for s in s_dcoms:
            assert s in s_coms

    def test_find_blocks_first_passing(self):
        youshallnotpass = [augeasparser.AugeasBlockNode(name="notpassing",
                                                        ancestor=self.block,
                                                        filepath="/path/to/whatever",
                                                        metadata=self.metadata)]
        youshallpass = [augeasparser.AugeasBlockNode(name=assertions.PASS,
                                                     ancestor=self.block,
                                                     filepath=assertions.PASS,
                                                     metadata=self.metadata)]
        find_blocks_primary = mock.MagicMock(return_value=youshallpass)
        find_blocks_secondary = mock.MagicMock(return_value=youshallnotpass)
        self.block.primary.find_blocks = find_blocks_primary
        self.block.secondary.find_blocks = find_blocks_secondary

        blocks = self.block.find_blocks("something")
        for block in blocks:
            try:
                assertions.assertEqual(block.primary, block.secondary)
            except AssertionError: # pragma: no cover
                self.fail("Assertion should have passed")
            assert assertions.isPassDirective(block.primary) is True
            assert assertions.isPassDirective(block.secondary) is False

    def test_find_blocks_second_passing(self):
        youshallnotpass = [augeasparser.AugeasBlockNode(name="notpassing",
                                                        ancestor=self.block,
                                                        filepath="/path/to/whatever",
                                                        metadata=self.metadata)]
        youshallpass = [augeasparser.AugeasBlockNode(name=assertions.PASS,
                                                     ancestor=self.block,
                                                     filepath=assertions.PASS,
                                                     metadata=self.metadata)]
        find_blocks_primary = mock.MagicMock(return_value=youshallnotpass)
        find_blocks_secondary = mock.MagicMock(return_value=youshallpass)
        self.block.primary.find_blocks = find_blocks_primary
        self.block.secondary.find_blocks = find_blocks_secondary

        blocks = self.block.find_blocks("something")
        for block in blocks:
            try:
                assertions.assertEqual(block.primary, block.secondary)
            except AssertionError: # pragma: no cover
                self.fail("Assertion should have passed")
            assert assertions.isPassDirective(block.primary) is False
            assert assertions.isPassDirective(block.secondary) is True

    def test_find_dirs_first_passing(self):
        notpassing = [augeasparser.AugeasDirectiveNode(name="notpassing",
                                                       ancestor=self.block,
                                                       filepath="/path/to/whatever",
                                                       metadata=self.metadata)]
        passing = [augeasparser.AugeasDirectiveNode(name=assertions.PASS,
                                                    ancestor=self.block,
                                                    filepath=assertions.PASS,
                                                    metadata=self.metadata)]
        find_dirs_primary = mock.MagicMock(return_value=passing)
        find_dirs_secondary = mock.MagicMock(return_value=notpassing)
        self.block.primary.find_directives = find_dirs_primary
        self.block.secondary.find_directives = find_dirs_secondary

        directives = self.block.find_directives("something")
        for directive in directives:
            try:
                assertions.assertEqual(directive.primary, directive.secondary)
            except AssertionError: # pragma: no cover
                self.fail("Assertion should have passed")
            assert assertions.isPassDirective(directive.primary) is True
            assert assertions.isPassDirective(directive.secondary) is False

    def test_find_dirs_second_passing(self):
        notpassing = [augeasparser.AugeasDirectiveNode(name="notpassing",
                                                       ancestor=self.block,
                                                       filepath="/path/to/whatever",
                                                       metadata=self.metadata)]
        passing = [augeasparser.AugeasDirectiveNode(name=assertions.PASS,
                                                    ancestor=self.block,
                                                    filepath=assertions.PASS,
                                                    metadata=self.metadata)]
        find_dirs_primary = mock.MagicMock(return_value=notpassing)
        find_dirs_secondary = mock.MagicMock(return_value=passing)
        self.block.primary.find_directives = find_dirs_primary
        self.block.secondary.find_directives = find_dirs_secondary

        directives = self.block.find_directives("something")
        for directive in directives:
            try:
                assertions.assertEqual(directive.primary, directive.secondary)
            except AssertionError: # pragma: no cover
                self.fail("Assertion should have passed")
            assert assertions.isPassDirective(directive.primary) is False
            assert assertions.isPassDirective(directive.secondary) is True

    def test_find_coms_first_passing(self):
        notpassing = [augeasparser.AugeasCommentNode(comment="notpassing",
                                                     ancestor=self.block,
                                                     filepath="/path/to/whatever",
                                                     metadata=self.metadata)]
        passing = [augeasparser.AugeasCommentNode(comment=assertions.PASS,
                                                  ancestor=self.block,
                                                  filepath=assertions.PASS,
                                                  metadata=self.metadata)]
        find_coms_primary = mock.MagicMock(return_value=passing)
        find_coms_secondary = mock.MagicMock(return_value=notpassing)
        self.block.primary.find_comments = find_coms_primary
        self.block.secondary.find_comments = find_coms_secondary

        comments = self.block.find_comments("something")
        for comment in comments:
            try:
                assertions.assertEqual(comment.primary, comment.secondary)
            except AssertionError: # pragma: no cover
                self.fail("Assertion should have passed")
            assert assertions.isPassComment(comment.primary) is True
            assert assertions.isPassComment(comment.secondary) is False

    def test_find_coms_second_passing(self):
        notpassing = [augeasparser.AugeasCommentNode(comment="notpassing",
                                                     ancestor=self.block,
                                                     filepath="/path/to/whatever",
                                                     metadata=self.metadata)]
        passing = [augeasparser.AugeasCommentNode(comment=assertions.PASS,
                                                  ancestor=self.block,
                                                  filepath=assertions.PASS,
                                                  metadata=self.metadata)]
        find_coms_primary = mock.MagicMock(return_value=notpassing)
        find_coms_secondary = mock.MagicMock(return_value=passing)
        self.block.primary.find_comments = find_coms_primary
        self.block.secondary.find_comments = find_coms_secondary

        comments = self.block.find_comments("something")
        for comment in comments:
            try:
                assertions.assertEqual(comment.primary, comment.secondary)
            except AssertionError: # pragma: no cover
                self.fail("Assertion should have passed")
            assert assertions.isPassComment(comment.primary) is False
            assert assertions.isPassComment(comment.secondary) is True

    def test_find_blocks_no_pass_equal(self):
        notpassing1 = [augeasparser.AugeasBlockNode(name="notpassing",
                                                    ancestor=self.block,
                                                    filepath="/path/to/whatever",
                                                    metadata=self.metadata)]
        notpassing2 = [augeasparser.AugeasBlockNode(name="notpassing",
                                                    ancestor=self.block,
                                                    filepath="/path/to/whatever",
                                                    metadata=self.metadata)]
        find_blocks_primary = mock.MagicMock(return_value=notpassing1)
        find_blocks_secondary = mock.MagicMock(return_value=notpassing2)
        self.block.primary.find_blocks = find_blocks_primary
        self.block.secondary.find_blocks = find_blocks_secondary

        blocks = self.block.find_blocks("anything")
        assert len(blocks) == 1
        for block in blocks:
            assert block.primary == block.secondary
            assert block.primary is not block.secondary

    def test_find_dirs_no_pass_equal(self):
        notpassing1 = [augeasparser.AugeasDirectiveNode(name="notpassing",
                                                        ancestor=self.block,
                                                        filepath="/path/to/whatever",
                                                        metadata=self.metadata)]
        notpassing2 = [augeasparser.AugeasDirectiveNode(name="notpassing",
                                                        ancestor=self.block,
                                                        filepath="/path/to/whatever",
                                                        metadata=self.metadata)]
        find_dirs_primary = mock.MagicMock(return_value=notpassing1)
        find_dirs_secondary = mock.MagicMock(return_value=notpassing2)
        self.block.primary.find_directives = find_dirs_primary
        self.block.secondary.find_directives = find_dirs_secondary

        directives = self.block.find_directives("anything")
        assert len(directives) == 1
        for directive in directives:
            assert directive.primary == directive.secondary
            assert directive.primary is not directive.secondary

    def test_find_comments_no_pass_equal(self):
        notpassing1 = [augeasparser.AugeasCommentNode(comment="notpassing",
                                                      ancestor=self.block,
                                                      filepath="/path/to/whatever",
                                                      metadata=self.metadata)]
        notpassing2 = [augeasparser.AugeasCommentNode(comment="notpassing",
                                                      ancestor=self.block,
                                                      filepath="/path/to/whatever",
                                                      metadata=self.metadata)]
        find_coms_primary = mock.MagicMock(return_value=notpassing1)
        find_coms_secondary = mock.MagicMock(return_value=notpassing2)
        self.block.primary.find_comments = find_coms_primary
        self.block.secondary.find_comments = find_coms_secondary

        comments = self.block.find_comments("anything")
        assert len(comments) == 1
        for comment in comments:
            assert comment.primary == comment.secondary
            assert comment.primary is not comment.secondary

    def test_find_blocks_no_pass_notequal(self):
        notpassing1 = [augeasparser.AugeasBlockNode(name="notpassing",
                                                    ancestor=self.block,
                                                    filepath="/path/to/whatever",
                                                    metadata=self.metadata)]
        notpassing2 = [augeasparser.AugeasBlockNode(name="different",
                                                    ancestor=self.block,
                                                    filepath="/path/to/whatever",
                                                    metadata=self.metadata)]
        find_blocks_primary = mock.MagicMock(return_value=notpassing1)
        find_blocks_secondary = mock.MagicMock(return_value=notpassing2)
        self.block.primary.find_blocks = find_blocks_primary
        self.block.secondary.find_blocks = find_blocks_secondary

        with pytest.raises(AssertionError):
            _ = self.block.find_blocks("anything")

    def test_parsernode_notequal(self):
        ne_block = augeasparser.AugeasBlockNode(name="different",
                                                ancestor=self.block,
                                                filepath="/path/to/whatever",
                                                metadata=self.metadata)
        ne_directive = augeasparser.AugeasDirectiveNode(name="different",
                                                        ancestor=self.block,
                                                        filepath="/path/to/whatever",
                                                        metadata=self.metadata)
        ne_comment = augeasparser.AugeasCommentNode(comment="different",
                                                    ancestor=self.block,
                                                    filepath="/path/to/whatever",
                                                    metadata=self.metadata)
        assert self.block != ne_block
        assert self.directive != ne_directive
        assert self.comment != ne_comment

    def test_parsed_paths(self):
        mock_p = mock.MagicMock(return_value=['/path/file.conf',
                                              '/another/path',
                                              '/path/other.conf'])
        mock_s = mock.MagicMock(return_value=['/path/*.conf', '/another/path'])
        self.block.primary.parsed_paths = mock_p
        self.block.secondary.parsed_paths = mock_s
        self.block.parsed_paths()
        assert mock_p.called is True
        assert mock_s.called is True

    def test_parsed_paths_error(self):
        mock_p = mock.MagicMock(return_value=['/path/file.conf'])
        mock_s = mock.MagicMock(return_value=['/path/*.conf', '/another/path'])
        self.block.primary.parsed_paths = mock_p
        self.block.secondary.parsed_paths = mock_s
        with pytest.raises(AssertionError):
            self.block.parsed_paths()

    def test_find_ancestors(self):
        primarymock = mock.MagicMock(return_value=[])
        secondarymock = mock.MagicMock(return_value=[])
        self.block.primary.find_ancestors = primarymock
        self.block.secondary.find_ancestors = secondarymock
        self.block.find_ancestors("anything")
        assert primarymock.called is True
        assert secondarymock.called is True
